#pylint: disable=line-too-long,too-many-lines,too-many-statements,anomalous-backslash-in-string
"""
Representation of an optimal power flow result.
"""

import logging
import os.path
from collections import OrderedDict

import numpy as np
import pandas as pd

from hynet import __version__ as hynet_version
from hynet.types_ import hynet_float_, BusType, BranchType, SolverType, SolverStatus
from hynet.utilities.base import Timer, truncate_with_ellipsis
from hynet.utilities.rank1approx import calc_rank1approx_rel_err

_log = logging.getLogger(__name__)


class OPFResult:
    """
    Result of an optimal power flow calculation.

    **Remark:** In the data frames below, the respective column for the dual
    variables of a type of constraint (e.g., voltage drop) is only present if
    at least one constraint of this constraint type appears in the problem
    formulation.

    Parameters
    ----------
    model : SystemModel
        Steady-state model for which the optimal power flow was calculated.
    empty : bool
        ``True`` if the object does not contain any result data and ``False``
        otherwise.
    solver : SolverInterface
        Solver object by which the result was obtained.
    solver_status : SolverStatus
        Status reported by the solver.
    solver_time : float
        Duration of the call to the solver in seconds.
    optimal_value : float
        Optimal objective value or ``numpy.nan`` if the solver failed.
    total_time : float or numpy.nan
        Total time for the OPF calculation, including loading of the data (if
        applicable), modeling, solving, and result assembly. If not provided,
        this time is set to ``numpy.nan``.
    is_physical : bool
        Unavailable if the result is empty and, otherwise, ``True`` if the
        power balance equations hold with reasonable accuracy.
    reconstruction_mse : float
        Unavailable if the result is empty and, otherwise, the mean squared
        error of the reconstructed bus voltages in case of a relaxation and
        ``numpy.nan`` otherwise.
    total_injection_cost : float
        Unavailable if the result is empty and, otherwise, the total injection
        cost in $/h.
    dynamic_losses : float
        Unavailable if the result is empty and, otherwise, the dynamic losses
        in MW.
    total_losses : float
        Unavailable if the result is empty and, otherwise, the total losses in
        MW (dynamic losses plus the static losses of the converters).

    bus : pandas.DataFrame, optional
        Unavailable if the result is empty and, otherwise, a data frame with
        the bus result data, indexed by the *bus ID*, which comprises the
        following columns:

        ``v``: (``hynet_complex_``)
            Bus voltage rms phasor (AC) or bus voltage magnitude (DC).
        ``s_shunt``: (``hynet_complex_``)
            Shunt apparent power in MVA. The real part constitutes the shunt
            losses in MW and the *negated* imaginary part constitutes the
            reactive power *injection*.
        ``dv_v_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the voltage lower bound in $/p.u..
        ``dv_v_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the voltage upper bound in $/p.u..
        ``dv_bal_p``: (``hynet_float_``)
            Dual variable or KKT multiplier of the active power balance
            constraint in $/MW. In case of exactness of the relaxation (or
            zero duality gap in the QCQP), these dual variables equal the
            locational marginal prices (LMPs) for active power, cf. [1]_.
        ``dv_bal_q``: (``hynet_float_``)
            Dual variable or KKT multiplier of the reactive power balance
            constraint in $/Mvar. In case of exactness of the relaxation (or
            zero duality gap in the QCQP), these dual variables equal the
            locational marginal prices (LMPs) for reactive power, cf. [1]_.
        ``bal_err``: (``hynet_complex_``)
            Power balance residual in MVA, i.e., the evaluation of the
            complex-valued power balance equation at the system state.
            Theoretically, this should be identical to zero, but due to a
            limited solver accuracy and/or inexactness of the relaxation it is
            only approximately zero. This residual supports the assessment of
            solution accuracy and validity.

    branch : pandas.DataFrame, optional
        Unavailable if the result is empty and, otherwise, a data frame with
        the branch result data, indexed by the *branch ID*, which comprises the
        following columns:

        ``s_src``: (``hynet_complex_``)
            Apparent power flow in MVA at the source bus (measured as a flow
            *into* the branch).
        ``s_dst``: (``hynet_complex_``)
            Apparent power flow in MVA at the destination bus (measured as a
            flow *into* the branch).
        ``i_src``: (``hynet_complex_``)
            Current flow in p.u. at the source bus (measured as a flow
            *into* the branch).
        ``i_dst``: (``hynet_complex_``)
            Current flow in p.u. at the destination bus (measured as a flow
            *into* the branch).
        ``v_drop``: (``hynet_float_``)
            Relative voltage magnitude drop from the source bus to the
            destination bus.
        ``angle_diff``: (``hynet_float_``)
            Bus voltage angle difference in degrees between the source and
            destination bus.
        ``effective_rating``: (``hynet_float_``)
            Ampacity in terms of a long-term MVA rating at the *actual* bus
            voltage. If no rating is available, it is set to ``numpy.nan``.
        ``rel_err``: (``hynet_float_``)
            Branch-related relative reconstruction error
            :math:`\kappa_k(V^\star)` as defined in equation (24) in [1]_ in
            case of a relaxed OPF or ``numpy.nan`` otherwise.
        ``dv_i_max_src``: (``hynet_float_``)
            Dual variable or KKT multiplier of the ampacity constraint at the
            source bus in $/p.u. or ``numpy.nan`` if unavailable.
        ``dv_i_max_dst``: (``hynet_float_``)
            Dual variable or KKT multiplier of the ampacity constraint at the
            destination bus in $/p.u. or ``numpy.nan`` if unavailable.
        ``dv_angle_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the angle difference lower bound
            constraint or ``numpy.nan`` if unavailable.
        ``dv_angle_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the angle difference upper bound
            constraint or ``numpy.nan`` if unavailable.
        ``dv_real_part``: (``hynet_float_``)
            Dual variable or KKT multiplier of the +/-90 degrees constraint
            on the angle difference (cf. equation (27) in [2]_) or
            ``numpy.nan`` if unavailable.
        ``dv_drop_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the voltage drop lower bound
            constraint or ``numpy.nan`` if unavailable.
        ``dv_drop_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the voltage drop upper bound
            constraint or ``numpy.nan`` if unavailable.

    converter : pandas.DataFrame
        Unavailable if the result is empty and, otherwise, a data frame with
        the converter result data, indexed by the *converter ID*, which
        comprises the following columns:

        ``p_src``: (``hynet_complex_``)
            Active power flow in MW at the source bus *into the converter*.
        ``p_dst``: (``hynet_complex_``)
            Active power flow in MW at the destination bus *into the
            converter*.
        ``q_src``: (``hynet_complex_``)
            Reactive power injection in Mvar at the source bus *into the grid*.
        ``q_dst``: (``hynet_complex_``)
            Reactive power injection in Mvar at the destination bus *into the
            grid*.
        ``dv_cap_src_p_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the active power lower bound
            of the capability region at the source bus in $/MW or
            ``numpy.nan`` if unavailable.
        ``dv_cap_src_q_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the reactive power lower bound
            of the capability region at the source bus in $/Mvar or
            ``numpy.nan`` if unavailable.
        ``dv_cap_src_p_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the active power upper bound
            of the capability region at the source bus in $/MW or
            ``numpy.nan`` if unavailable.
        ``dv_cap_src_q_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the reactive power upper bound
            of the capability region at the source bus in $/Mvar or
            ``numpy.nan`` if unavailable.
        ``dv_cap_src_lt``: (``hynet_float_``)
            Dual variable or KKT multiplier of the left-top half-space of
            of the capability region at the source bus in $/MVA or
            ``numpy.nan`` if unavailable.
        ``dv_cap_src_rt``: (``hynet_float_``)
            Dual variable or KKT multiplier of the right-top half-space of
            of the capability region at the source bus in $/MVA or
            ``numpy.nan`` if unavailable.
        ``dv_cap_src_lb``: (``hynet_float_``)
            Dual variable or KKT multiplier of the left-bottom half-space of
            of the capability region at the source bus in $/MVA or
            ``numpy.nan`` if unavailable.
        ``dv_cap_src_rb``: (``hynet_float_``)
            Dual variable or KKT multiplier of the right-bottom half-space of
            of the capability region at the source bus in $/MVA or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_p_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the active power lower bound
            of the capability region at the destination bus in $/MW or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_q_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the reactive power lower bound
            of the capability region at the destination bus in $/Mvar or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_p_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the active power upper bound
            of the capability region at the destination bus in $/MW or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_q_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the reactive power upper bound
            of the capability region at the destination bus in $/Mvar or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_lt``: (``hynet_float_``)
            Dual variable or KKT multiplier of the left-top half-space of
            of the capability region at the destination bus in $/MVA or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_rt``: (``hynet_float_``)
            Dual variable or KKT multiplier of the right-top half-space of
            of the capability region at the destination bus in $/MVA or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_lb``: (``hynet_float_``)
            Dual variable or KKT multiplier of the left-bottom half-space of
            of the capability region at the destination bus in $/MVA or
            ``numpy.nan`` if unavailable.
        ``dv_cap_dst_rb``: (``hynet_float_``)
            Dual variable or KKT multiplier of the right-bottom half-space of
            of the capability region at the destination bus in $/MVA or
            ``numpy.nan`` if unavailable.

    injector : pandas.DataFrame
        Unavailable if the result is empty and, otherwise, a data frame with
        the injector result data, indexed by the *injector ID*, which comprises
        the following columns:

        ``s``: (``hynet_complex_``)
            Apparent power injection in MVA.
        ``cost_p``: (``hynet_float_``)
            Cost of the active power injection in dollars or ``numpy.nan`` if
            no cost function was provided.
        ``cost_q``: (``hynet_float_``)
            Cost of the reactive power injection in dollars or ``numpy.nan`` if
            no cost function was provided.
        ``dv_cap_p_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the active power lower bound
            of the capability region in $/MW or ``numpy.nan`` if unavailable.
        ``dv_cap_q_min``: (``hynet_float_``)
            Dual variable or KKT multiplier of the reactive power lower bound
            of the capability region in $/Mvar or ``numpy.nan`` if unavailable.
        ``dv_cap_p_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the active power upper bound
            of the capability region in $/MW or ``numpy.nan`` if unavailable.
        ``dv_cap_q_max``: (``hynet_float_``)
            Dual variable or KKT multiplier of the reactive power upper bound
            of the capability region in $/Mvar or ``numpy.nan`` if unavailable.
        ``dv_cap_lt``: (``hynet_float_``)
            Dual variable or KKT multiplier of the left-top half-space of
            the capability region in $/MVA or ``numpy.nan`` if unavailable.
        ``dv_cap_rt``: (``hynet_float_``)
            Dual variable or KKT multiplier of the right-top half-space of
            the capability region in $/MVA or ``numpy.nan`` if unavailable.
        ``dv_cap_lb``: (``hynet_float_``)
            Dual variable or KKT multiplier of the left-bottom half-space of
            the capability region in $/MVA or ``numpy.nan`` if unavailable.
        ``dv_cap_rb``: (``hynet_float_``)
            Dual variable or KKT multiplier of the right-bottom half-space of
            the capability region in $/MVA or ``numpy.nan`` if unavailable.

    References
    ----------
    .. [1] M. Hotz and W. Utschick, "The Hybrid Transmission Grid Architecture:
           Benefits in Nodal Pricing," in IEEE Trans. Power Systems, vol. 33,
           no. 2, pp. 1431-1442, Mar. 2018.
    .. [2] M. Hotz and W. Utschick, "A Hybrid Transmission Grid Architecture
           Enabling Efficient Optimal Power Flow," in IEEE Trans. Power
           Systems, vol. 31, no. 6, pp. 4504-4516, Nov. 2016.
    """

    # Result printing: Threshold on zero values to be replaced by a dash
    _zero_thres = 1e-8

    # Result printing: Tolerance factor for activity indicator w.r.t.
    # constraint interval
    _tol_active = 0.005

    # Result printing: Ratio w.r.t. to rating for indication of high loading
    _high_loading = 0.9

    # Physical solution: Threshold on ratio of power balance error to largest
    # injector/load to consider the power balance as sufficiently accurate.
    _bal_err_rel_thres = 1e-3

    def __init__(self, model, qcqp_result, total_time=np.nan):
        """
        Create an OPF result object.

        Parameters
        ----------
        model : hynet.model.steady_state.SystemModel
            System model for which the OPF was performed.
        qcqp_result : hynet.qcqp.result.QCQPResult
            Solution of the OPF QCQP.
        total_time : .hynet_float_, optional
            Total time for solving the OPF, cf. hynet.opf.calc.calc_opf.
        """
        timer = Timer()
        self.model = model
        self.total_time = total_time

        # REMARK: Only relevant data of the QCQP result object is extracted
        # here as storing it in its entirety would bloat the OPF result object
        # due to the constraint objects in the contained QCQP object, which
        # exhibit a significant memory consumption.
        self.empty = qcqp_result.empty
        self.solver = qcqp_result.solver
        self.solver_status = qcqp_result.solver_status
        self.solver_time = qcqp_result.solver_time
        self.optimal_value = qcqp_result.optimal_value
        self._optimizer = qcqp_result.optimizer

        if qcqp_result.empty:
            return

        # Bus-related state data
        s_shunt = self.model.calc_shunt_apparent_power(self._optimizer)

        self.bus = pd.DataFrame(OrderedDict(
            [('v', self._optimizer.v),
             ('s_shunt', s_shunt)
             ]))
        self.bus.index = self.model.scenario.bus.index

        # Branch-related state data
        i_src, i_dst, s_src, s_dst = self.model.calc_branch_flow(self._optimizer)
        v_drop = self.model.calc_branch_voltage_drop(self._optimizer)
        angle_diff = self.model.calc_branch_angle_difference(self._optimizer)
        effective_rating = self.model.calc_branch_effective_rating(self._optimizer)

        self.branch = pd.DataFrame(OrderedDict(
            [('s_src', s_src),
             ('s_dst', s_dst),
             ('i_src', i_src),
             ('i_dst', i_dst),
             ('v_drop', v_drop),
             ('angle_diff', angle_diff),
             ('effective_rating', effective_rating)
             ]))
        self.branch.index = self.model.scenario.branch.index

        # Converter-related state data
        s_src, s_dst = self.model.calc_converter_flow(self._optimizer)

        self.converter = pd.DataFrame(OrderedDict(
            [('p_src', s_src.real),
             ('p_dst', s_dst.real),
             ('q_src', -s_src.imag),
             ('q_dst', -s_dst.imag)
             ]))
        self.converter.index = self.model.scenario.converter.index

        # Injector-related state data
        cost_p, cost_q = self.model.calc_injection_costs(self._optimizer)

        self.injector = pd.DataFrame(OrderedDict(
            [('s', self._optimizer.s[0::2] + 1j*self._optimizer.s[1::2]),
             ('cost_p', cost_p),
             ('cost_q', cost_q)
             ]))
        self.injector.index = self.model.scenario.injector.index

        # Mean squared and branch-related reconstruction error
        self.reconstruction_mse = qcqp_result.reconstruction_mse
        if qcqp_result.V is not None:
            edges = (self.model.scenario.e_src.values,
                     self.model.scenario.e_dst.values)
            rel_err = calc_rank1approx_rel_err(qcqp_result.V,
                                               self._optimizer.v,
                                               edges)
        else:
            rel_err = np.nan * np.ones(self.model.scenario.num_branches,
                                       dtype=hynet_float_)

        self.branch.loc[:, 'rel_err'] = pd.Series(rel_err,
                                                  index=self.branch.index)

        # Dual variables of the box constraints
        #
        # REMARK 1: As the QCQP considers the variables in p.u., the dual
        # variables must be scaled accordingly to capture the sensitivity in
        # $/MW and $/Mvar, respectively.
        #
        # REMARK 2: The active power limits of the converter capability regions
        # are implemented via inequalities, see also ``ConverterCapRegion``,
        # and thus their dual variables are not extracted here.
        self.injector['dv_cap_p_min'] = qcqp_result.dv_lb.s[0::2]
        self.injector['dv_cap_q_min'] = qcqp_result.dv_lb.s[1::2]

        self.injector['dv_cap_p_max'] = qcqp_result.dv_ub.s[0::2]
        self.injector['dv_cap_q_max'] = qcqp_result.dv_ub.s[1::2]

        self.converter['dv_cap_src_q_min'] = qcqp_result.dv_lb.f[2::4]
        self.converter['dv_cap_dst_q_min'] = qcqp_result.dv_lb.f[3::4]

        self.converter['dv_cap_src_q_max'] = qcqp_result.dv_ub.f[2::4]
        self.converter['dv_cap_dst_q_max'] = qcqp_result.dv_ub.f[3::4]

        # Dual variables and constraint function values
        timer.interval()
        qcqp_result.get_result_tables(tables={'bus': self.bus,
                                              'branch': self.branch,
                                              'converter': self.converter,
                                              'injector': self.injector},
                                      dual_prefix='dv_',
                                      value_prefix='cv_')
        _log.debug("Result table creation ({:.3f} sec.)"
                   .format(timer.interval()))

        self.bus.loc[:, 'bal_err'] = self.bus['cv_bal_p'] + 1j*self.bus['cv_bal_q']
        self.bus.drop(['cv_bal_p', 'cv_bal_q'], axis='columns', inplace=True)

        # Total injection cost
        self.total_injection_cost = 0
        if not self.injector['cost_p'].isnull().all():
            self.total_injection_cost += self.injector['cost_p'].sum()
        if not self.injector['cost_q'].isnull().all():
            self.total_injection_cost += self.injector['cost_q'].sum()

        # Total losses
        self.dynamic_losses = \
            self.model.calc_dynamic_losses(self._optimizer)
        self.total_losses = self.dynamic_losses
        if not self.model.scenario.converter.empty:
            self.total_losses += self.model.scenario.converter['loss_fix'].sum()

        # Physically valid solution: Do the state variables satisfy the power
        # balance equations up to a sufficient accuracy?
        max_bal_err = self.bus.loc[:, 'bal_err'].abs().max()
        max_inj = np.max([self.model.scenario.bus['load'].abs().max(),
                          self.injector['s'].abs().max()])
        self.is_physical = max_bal_err / max_inj <= self._bal_err_rel_thres

        # Update the total time with the duration of the OPF result creation
        self.total_time += timer.total()

        _log.debug("OPF result creation (total) ({:.3f} sec.)"
                   .format(timer.total()))

    @property
    def num_buses(self):
        """Return the number of buses."""
        return 0 if self.empty else len(self.bus.index)

    @property
    def num_branches(self):
        """Return the number of branches."""
        return 0 if self.empty else len(self.branch.index)

    @property
    def num_converters(self):
        """Return the number of converters."""
        return 0 if self.empty else len(self.converter.index)

    @property
    def num_injectors(self):
        """Return the number of injectors."""
        return 0 if self.empty else len(self.injector.index)

    @property
    def scenario(self):
        """Return the scenario data of the system model."""
        return self.model.scenario

    def __repr__(self):
        """Return a summary of the OPF result."""
        t = ""
        t += self._get_header()
        if self.empty:
            t += self._get_termination_summary()
        else:
            t += "|> Grid Information " + "-"*58 + "<|\n"
            t += "|" + " "*78 + "|\n"
            t += self._get_grid_summary()
            t += "|" + " "*78 + "|\n"
            t += "|> Results " + "-"*67 + "<|\n"
            t += "|" + " "*78 + "|\n"
            t += self._get_result_summary()
            t += "|" + " "*78 + "|\n"
            t += "|> Solver Information " + "-"*56 + "<|\n"
            t += "|" + " "*78 + "|\n"
            t += self._get_solver_info()
            t += "|" + " "*78 + "|\n"
            t += "+" + "-"*78 + "+\n"
        return t

    @property
    def details(self):
        """
        Return a formatted string with details of the OPF result.

        The returned string contains a formatted table for all major entity
        types, which is hopefully mostly self-explaining. In the very left or
        right of a column, there may be an indicator:

        +-----------+-------------------------------------------------------+
        | Indicator | Meaning                                               |
        +===========+=======================================================+
        |   ``R``   | Reference bus in the respective subgrid.              |
        +-----------+-------------------------------------------------------+
        |   ``=``   | The bus is a DC bus. If there is no indicator, the    |
        |           | bus is an AC bus.                                     |
        +-----------+-------------------------------------------------------+
        |   ``*``   | A limit on the respective quantity is active.         |
        +-----------+-------------------------------------------------------+
        |   ``>``   | The branch is highly loaded, i.e., the flow is 90% or |
        |           | more of the effective rating.                         |
        +-----------+-------------------------------------------------------+
        |   ``T``   | The branch is a transformer. If there is no           |
        |           | indicator, the branch is a line/cable.                |
        +-----------+-------------------------------------------------------+
        """
        if self.empty:
            return self._get_termination_summary()
        if self.bus.index.max() > 99999 or self.bus.index.min() < -9999:
            raise RuntimeError("This formatted output only supports bus IDs "
                               "comprising 5 characters or less.")
        t = ""
        t += self._get_bus_details()
        t += self._get_branch_details()
        t += self._get_converter_details()
        t += self._get_injector_details()
        return t

    #==========================================================================
    #
    # CAUTION: Below is only some ugly string formatting code.
    #
    #                         You have been warned! ;P
    #
    #==========================================================================

    def _get_header(self):
        scenario = self.scenario
        t = ""
        t += "\n+" + "-"*78 + "+\n"
        t += "| OPTIMAL POWER FLOW {:>57s} |\n"\
            .format("hynet ~ version " + hynet_version)
        t += "|" + " "*78 + "|\n"
        t += "|> Data Source " + "-"*63 + "<|\n"
        t += "|" + " "*78 + "|\n"
        t += "| Grid:       {:<64s} |\n"\
            .format(truncate_with_ellipsis(scenario.grid_name + " ("
                + os.path.basename(scenario.database_uri) + ")", 64))
        t += "| Scenario:   {:<64s} |\n"\
            .format(truncate_with_ellipsis("{0:s} @ {1:s} (id={2:s})"
                                           .format(scenario.name,
                                                   scenario.get_time_string(),
                                                   str(scenario.id)),
                                           64))
        t += "|" + " "*78 + "|\n"
        return t

    def _get_termination_summary(self):
        t = ""
        t += "+" + "-"*78 + "+\n"
        t += "| Solver terminated with status {:<46s} |\n"\
            .format("'" + self.solver_status.name + "'")
        t += "+" + "-"*78 + "+\n"
        return t

    def _get_grid_summary(self):
        bus = self.scenario.bus
        branch = self.scenario.branch
        converter = self.scenario.converter
        injector = self.scenario.injector

        line_bus_type = \
            bus.loc[branch.loc[branch['type'] == BranchType.LINE, 'src'], 'type'].values

        conv_src_bus_type = bus.loc[converter['src'], 'type'].values
        conv_dst_bus_type = bus.loc[converter['dst'], 'type'].values
        conv_b2b_bus_type = conv_src_bus_type[conv_src_bus_type == conv_dst_bus_type]

        num_conventional = sum(x.is_conventional() for x in injector['type'])
        num_renewable = sum(x.is_renewable() for x in injector['type'])
        num_prosumer = sum(x.is_prosumer() for x in injector['type'])
        num_load = sum(x.is_load() for x in injector['type'])
        num_compensation = sum(x.is_compensation() for x in injector['type'])

        p_capacity = sum(x.p_max for x in injector['cap'])
        p_tot_load = bus['load'].real.sum()

        t = ""
        t += "| Topology: {:>9s} |{:7d} AC subgrids|{:6d} DC subgrids|{:5d} islands    |\n"\
            .format("acyclic" if self.model.is_acyclic else "meshed",
                    len(self.model.ac_subgrids),
                    len(self.model.dc_subgrids),
                    len(self.model.islands) - 1)
        t += "| Buses:     {:8d} |{:7d} AC buses   |{:6d} DC buses   |{:5d} with shunt |\n"\
            .format(len(bus.index),
                    (bus['type'] == BusType.AC).sum(),
                    (bus['type'] == BusType.DC).sum(),
                    (bus['y_tld'].abs() > 0).sum())
        t += "| Branches:  {:8d} |{:7d} AC lines   |{:6d} DC lines   |{:5d} transform. |\n"\
            .format(len(branch.index),
                    (line_bus_type == BusType.AC).sum(),
                    (line_bus_type == BusType.DC).sum(),
                    (branch['type'].values == BranchType.TRANSFORMER).sum())
        t += "| Converters:{:8d} |{:7d} AC/DC      |{:6d} AC/AC      |{:5d} DC/DC      |\n"\
            .format(len(converter.index),
                    (conv_src_bus_type != conv_dst_bus_type).sum(),
                    (conv_b2b_bus_type == BusType.AC).sum(),
                    (conv_b2b_bus_type == BusType.DC).sum())
        t += "| Injectors: {:8d} |{:7d} convent.   |{:6d} disp. load |{:5d} compensat. |\n"\
            .format(len(injector.index),
                    num_conventional,
                    num_load,
                    num_compensation)
        t += "|" + " "*21 + "|{:7d} renewable  |{:6d} prosumer   |                 |\n"\
            .format(num_renewable,
                    num_prosumer)
        t += "|" + " "*78 + "|\n"
        t += "| Injection:{:11.1f} MW /{:8.1f} Mvar | Min.: {:9.1f} MW /{:9.1f} Mvar |\n"\
            .format(p_capacity,
                    sum(x.q_max for x in injector['cap']),
                    sum(x.p_min for x in injector['cap']),
                    sum(x.q_min for x in injector['cap']))
        t += "| Total load:{:10.1f} MW /{:8.1f} Mvar | Loading:{:9.2%} of P-capacity    |\n"\
            .format(p_tot_load,
                    bus['load'].imag.sum(),
                    p_tot_load/p_capacity)
        return t

    def _get_result_summary(self):
        p_capacity = sum(x.p_max for x in self.scenario.injector['cap'])

        converter_p_sum = 0
        converter_q_sum = 0
        if not self.scenario.converter.empty:
            converter_p_sum += self.scenario.converter['loss_fix'].sum()
            converter_p_sum += self.converter['p_src'].sum()
            converter_p_sum += self.converter['p_dst'].sum()
            converter_q_sum += self.converter['q_src'].sum()
            converter_q_sum += self.converter['q_dst'].sum()

        v_abs = np.abs(self.bus['v'])
        v_angle = np.angle(self.bus['v']) * 180/np.pi

        dv_bal_p_min = self.bus['dv_bal_p'].min()
        dv_bal_p_max = self.bus['dv_bal_p'].max()
        dv_bal_q_min = self.bus['dv_bal_q'].min()
        dv_bal_q_max = self.bus['dv_bal_q'].max()

        t = ""
        t += "| Injection:{:11.1f} MW /{:8.1f} Mvar | Inj. cost:{:17.3f} k$/h   |\n"\
            .format(self.injector['s'].real.sum(),
                    self.injector['s'].imag.sum(),
                    self.total_injection_cost/1e3)
        t += "| Total loss:{:10.1f} MW /{:6.2%} P-cap. | Shunt loss:{:14.1f} MW       |\n"\
            .format(self.total_losses,
                    self.total_losses/p_capacity,
                    self.bus['s_shunt'].real.sum())
        t += "| Branch loss:{:9.1f} MW                | Shunt Q-sum:{:13.1f} Mvar     |\n"\
            .format(self.branch['s_src'].real.sum() + self.branch['s_dst'].real.sum(),
                    -self.bus['s_shunt'].imag.sum())
        t += "| Conv. loss:{:10.1f} MW                | Converter Q-sum:{:9.1f} Mvar     |\n"\
            .format(converter_p_sum, converter_q_sum)
        t += "|" + " "*78 + "|\n"
        t += "| Voltage mag.:{:10.3f} to{:10.3f} p.u."\
            .format(v_abs.min(), v_abs.max()) + " "*36 + "|\n"
        t += "| Voltage angle:{:8.2f}  to{:9.2f} deg."\
            .format(v_angle.min(), v_angle.max()) + " "*37 + "|\n"
        t += "| P-bal. dual:{:>10s}  to{:>9s} $/MWh{:36s}|\n"\
            .format("{:10.2f}".format(dv_bal_p_min) if not np.isnan(dv_bal_p_min) else "-  ",
                    "{:9.2f}".format(dv_bal_p_max) if not np.isnan(dv_bal_p_max) else "-  ", "")
        t += "| Q-bal. dual:{:>10s}  to{:>9s} $/Mvarh{:34s}|\n"\
            .format("{:10.2f}".format(dv_bal_q_min) if not np.isnan(dv_bal_q_min) else "-  ",
                    "{:9.2f}".format(dv_bal_q_max) if not np.isnan(dv_bal_q_max) else "-  ", "")
        return t

    def _get_solver_info(self):
        solver_string = str(self.solver)
        if self.solver.param:
            solver_string += ' / '
            solver_string += ', '.join([key + '=' + str(value)
                                        for key, value
                                        in self.solver.param.items()])

        bal_err = self.bus['bal_err'].abs()

        if np.isnan(self.reconstruction_mse):
            rec_mse = "-"
        else:
            rec_mse = "{:9.3e} mean squared error".format(self.reconstruction_mse)
            if self.model.is_acyclic:
                rec_mse += "  <-  HYBRID ARCHITECTURE"

        t = ""
        t += "| Solver:             {:<56s} |\n"\
            .format(truncate_with_ellipsis(solver_string, 56))
        t += "| Status:             {:<56s} |\n"\
            .format(self.solver_status.name.lower())
        t += "| Time:               {:<56s} |\n"\
            .format('{:.2f} sec. in the solver'.format(self.solver_time)
                    + ('' if np.isnan(self.total_time) else
                       ' / {:.2f} sec. in total'.format(self.total_time)))
        t += "| Objective value:   {:10.3e} $/h"\
            .format(self.optimal_value) + " "*44 + "|\n"
        t += "| Loss penalty:      {:10.3e} $/h {:<42s} |\n"\
            .format(self.dynamic_losses * self.scenario.loss_price,
                    "({:.2f} MW * {:.3f} $/MWh)"\
                        .format(self.dynamic_losses, self.scenario.loss_price))
        t += "| Power balance:      {:9.3e} ({:9.3e}) mean (max.) absolute " \
             "error in MVA".format(bal_err.mean(), bal_err.max()) + "  |\n"
        t += "| Reconstruction:     {:<56s} |\n".format(rec_mse)
        if not self.is_physical:
            t += "|" + " "*78 + "|\n"
            t += "|{:^78s}|\n".format("WARNING: THE POWER BALANCE EQUATION IS VIOLATED!")
            if self.solver_status != SolverStatus.SOLVED:
                t += "|{:^78s}|\n".format("The solver encountered some issues. "
                                          "Try other settings/solvers.")
            elif self.solver.type != SolverType.QCQP:
                if self.model.is_acyclic:
                    t += "|{:^78s}|\n".format("A pathological price profile emerged. "
                                              "Try increasing the loss price.")
                elif self.solver.type == SolverType.SOCR:
                    t += "|{:^78s}|\n".format("The second-order cone relaxation is inexact. "
                                              "Try an SDR or QCQP solver.")
                else:
                    t += "|{:^78s}|\n".format("The semidefinite relaxation is inexact. "
                                              "Resort to a QCQP solver.")
        return t

    def _check_activity(self, value, min_, max_):
        if min_ is None and max_ is not None:
            min_ = np.nan * max_
        elif max_ is None and min_ is not None:
            max_ = np.nan * min_
        tolerance = pd.Series(False, index=max_.index)
        for id_ in tolerance.index:
            val = value.at[id_]
            lb = min_.at[id_]
            ub = max_.at[id_]
            lb_present = ~np.isnan(lb)
            ub_present = ~np.isnan(ub)
            if lb_present and ub_present:
                tol = self._tol_active * (ub - lb) + self._zero_thres
                tolerance.at[id_] = (val <= lb + tol) | (val >= ub - tol)
            elif lb_present:
                tol = abs(self._tol_active * lb) + self._zero_thres
                tolerance.at[id_] = (val <= lb + tol)
            elif ub_present:
                tol = abs(self._tol_active * ub) + self._zero_thres
                tolerance.at[id_] = (val >= ub - tol)
        return tolerance

    def _check_cap_region_box_active(self, value, cap):
        p = pd.Series(value.real, index=value.index)
        p_min = pd.Series([x.p_min for x in cap], index=cap.index)
        p_max = pd.Series([x.p_max for x in cap], index=cap.index)
        p_active = self._check_activity(p, p_min, p_max)

        q = pd.Series(value.imag, index=value.index)
        q_min = pd.Series([x.q_min for x in cap], index=cap.index)
        q_max = pd.Series([x.q_max for x in cap], index=cap.index)
        q_active = self._check_activity(q, q_min, q_max)
        return p_active, q_active

    def _get_bus_details(self):
        t = ""
        t += "\n+" + "-"*78 + "+\n"
        t += "| BUS RESULT" + " "*67 + "|\n"
        t += "+-------+---------------+-------------------+--------------+-------------------+\n"
        t += "|       |    Voltage    |       Load        |    Shunt     |  Dual Variables   |\n"
        t += "|       +-------+-------+---------+---------+------+-------+---------+---------+\n"
        t += "|       |  Mag. | Phase |    P    |    Q    | Loss | Q-Inj.|  P-bal. |  Q-bal. |\n"
        t += "|  ID   |  (pu) | (deg) |   (MW)  |  (Mvar) | (MW) | (Mvar)| ($/MWh) |($/Mvarh)|\n"
        t += "+-------+-------+-------+---------+---------+------+-------+---------+---------+\n"
        t = [t]
        scr_bus = self.scenario.bus
        res_bus = self.bus
        dc_bus = scr_bus['type'] == BusType.DC
        v_abs = res_bus['v'].abs()
        v_abs_active = self._check_activity(v_abs, scr_bus['v_min'], scr_bus['v_max'])
        v_angle = pd.Series(np.angle(res_bus['v']) * 180/np.pi, index=res_bus.index)
        load_p = pd.Series(scr_bus['load'].real, index=scr_bus.index)
        load_q = pd.Series(scr_bus['load'].imag, index=scr_bus.index)
        shunt_p = pd.Series(res_bus['s_shunt'].real, index=res_bus.index)
        shunt_q = pd.Series(res_bus['s_shunt'].imag, index=res_bus.index)
        dv_bal_p = res_bus['dv_bal_p']
        dv_bal_q = res_bus['dv_bal_q']
        for id_ in res_bus.index:
            r = "|"
            r += "R" if scr_bus.at[id_, 'ref'] else " "
            r += "{:5d}".format(id_)
            r += "=" if dc_bus.at[id_] else " "
            r += "|"
            r += "*" if v_abs_active.at[id_] else " "
            r += "{:6.3f}".format(v_abs.at[id_])
            r += "|"
            if dc_bus.at[id_] and np.abs(v_angle.at[id_]) < self._zero_thres:
                r += "   -   "
            else:
                r += "{:7.3f}".format(v_angle.at[id_])
            r += "|"
            if load_p.at[id_] == 0:
                r += "      -  "
            else:
                r += "{:9.2f}".format(load_p.at[id_])
            r += "|"
            if load_q.at[id_] == 0:
                r += "      -  "
            else:
                r += "{:9.2f}".format(load_q.at[id_])
            r += "|"
            if shunt_p.at[id_] == 0:
                r += "  -   "
            else:
                r += "{:6.3f}".format(shunt_p.at[id_])
            r += "|"
            if shunt_q.at[id_] == 0:
                r += "   -   "
            else:
                r += "{:7.3f}".format(-shunt_q.at[id_])
            r += "|"
            if np.isnan(dv_bal_p.at[id_]):
                r += "     -   "
            else:
                r += "{:9.3f}".format(dv_bal_p.at[id_])
            r += "|"
            if np.isnan(dv_bal_q.at[id_]):
                r += "     -   "
            else:
                r += "{:9.3f}".format(dv_bal_q.at[id_])
            r += "|"
            r += "\n"
            t.append(r)
        t = "".join(t)
        t += "+-------+-------+-------+---------+---------+------+-------+---------+---------+\n"
        return t

    def _get_branch_details(self):
        t = ""
        t += "\n+" + "-"*78 + "+\n"
        t += "| BRANCH RESULT" + " "*64 + "|\n"
        t += "+-------+-------------+-------------------+-----------------+--------+---------+\n"
        t += "|       |  Terminals  |    Power Flow     |     Voltage     | Losses | Reconst.|\n"
        t += "|       +------+------+----------+--------+--------+--------+--------+---------+\n"
        t += "|       | Src. | Dst. | |S_src|  | PF_src |  Drop  |Angle D.|  Dyn.  |  Kappa  |\n"
        t += "|  ID   | Bus  | Bus  |  (MVA)   |        |  (%)   | (deg)  |  (MW)  |         |\n"
        t += "+-------+------+------+----------+--------+--------+--------+--------+---------+\n"
        scr_branch = self.scenario.branch
        res_branch = self.branch

        if res_branch.empty:
            t += "| No branches present" + " "*58 + "|\n"
            t += "+" + "-"*78 + "+\n"
            return t

        t = [t]
        transformer = scr_branch['type'] == BranchType.TRANSFORMER
        src = scr_branch['src']
        dst = scr_branch['dst']
        dc_bus = self.scenario.bus['type'] == BusType.DC
        S_src = res_branch['s_src']
        S_dst = res_branch['s_dst']
        I_src_abs = res_branch['i_src'].abs()
        I_dst_abs = res_branch['i_dst'].abs()
        I_max = scr_branch['rating'] / self.scenario.base_mva
        congested = self._check_activity(I_src_abs, None, I_max) \
                  | self._check_activity(I_dst_abs, None, I_max)
        highly_loaded = (I_src_abs >= self._high_loading * I_max)\
                      | (I_dst_abs >= self._high_loading * I_max)
        drop = res_branch['v_drop'] * 100
        drop_active = self._check_activity(drop,
                                           scr_branch['drop_min'],
                                           scr_branch['drop_max'])
        angle_diff = res_branch['angle_diff']
        angle_diff_active = self._check_activity(angle_diff,
                                                 scr_branch['angle_min'],
                                                 scr_branch['angle_max'])
        loss_dyn = pd.Series((S_src + S_dst).real, index=res_branch.index)
        kappa = res_branch['rel_err']
        for id_ in res_branch.index:
            r = "|"
            r += "{:6d}".format(id_)
            r += "T" if transformer.at[id_] else " "
            r += "|"
            r += "{:5d}".format(src.at[id_])
            r += "=" if dc_bus.loc[src.at[id_]] else " "
            r += "|"
            r += "{:5d}".format(dst.at[id_])
            r += "=" if dc_bus.loc[dst.at[id_]] else " "
            r += "|"
            r += "*" if congested.at[id_] else ">" if highly_loaded.at[id_] else " "
            r += "{:9.2f}".format(np.abs(S_src.at[id_]))
            r += "|"
            if np.abs(S_src.at[id_]) < self._zero_thres:
                r += "    -   "
            else:
                r += "{:8.3f}".format(np.real(S_src.at[id_]) / np.abs(S_src.at[id_]))
            r += "|"
            r += "*" if drop_active.at[id_] else " "
            r += "{:7.3f}".format(drop.at[id_])
            r += "|"
            r += "*" if angle_diff_active.at[id_] else " "
            if dc_bus.loc[src.at[id_]] and np.abs(angle_diff.at[id_]) < self._zero_thres:
                r += "   -   "
            else:
                r += "{:7.3f}".format(angle_diff.at[id_])
            r += "|"
            r += "{:8.3f}".format(loss_dyn.at[id_])
            r += "|"
            if np.isnan(kappa.at[id_]):
                r += "     -   "
            else:
                r += "{:9.2e}".format(kappa.at[id_])
            r += "|"
            r += "\n"
            t.append(r)
        t = "".join(t)
        t += "+-------+------+------+----------+--------+--------+--------+--------+---------+\n"
        return t

    def _get_converter_details(self):
        t = ""
        t += "\n+" + "-"*78 + "+\n"
        t += "| CONVERTER RESULT" + " "*61 + "|\n"
        t += "+-------+-------------+---------------------+-----------------+----------------+\n"
        t += "|       |  Terminals  |     Power Flow      |Reactive Pwr Inj.|     Losses     |\n"
        t += "|       +------+------+----------+----------+--------+--------+--------+-------+\n"
        t += "|       | Src. | Dst. |  P_src   |  P_dst   | Q_src  | Q_dst  |  Dyn.  |  Fix  |\n"
        t += "|  ID   | Bus  | Bus  |   (MW)   |   (MW)   | (Mvar) | (Mvar) |  (MW)  |  (MW) |\n"
        t += "+-------+------+------+----------+----------+--------+--------+--------+-------+\n"
        scr_converter = self.scenario.converter
        res_converter = self.converter

        if res_converter.empty:
            t += "| No converters present" + " "*56 + "|\n"
            t += "+" + "-"*78 + "+\n"
            return t

        t = [t]
        src = scr_converter['src']
        dst = scr_converter['dst']
        dc_bus = self.scenario.bus['type'] == BusType.DC

        P_src = res_converter['p_src']
        P_dst = res_converter['p_dst']

        Q_src = res_converter['q_src']
        Q_dst = res_converter['q_dst']

        (P_src_active, _) = \
            self._check_cap_region_box_active(P_src + 1j*Q_src,
                                              scr_converter['cap_src'])
        (P_dst_active, _) = \
            self._check_cap_region_box_active(P_dst + 1j*Q_dst,
                                              scr_converter['cap_dst'])

        loss_dyn = P_src + P_dst
        loss_fix = scr_converter['loss_fix']

        for id_ in res_converter.index:
            r = "|"
            r += "{:6d}".format(id_)
            r += " |"
            r += "{:5d}".format(src.at[id_])
            r += "=" if dc_bus.loc[src.at[id_]] else " "
            r += "|"
            r += "{:5d}".format(dst.at[id_])
            r += "=" if dc_bus.loc[dst.at[id_]] else " "
            r += "|"
            r += "*" if P_src_active.at[id_] else " "
            r += "{:9.2f}".format(P_src.at[id_])
            r += "|"
            r += "*" if P_dst_active.at[id_] else " "
            r += "{:9.2f}".format(P_dst.at[id_])
            r += "|"
            if dc_bus.loc[src.at[id_]] and np.abs(Q_src.at[id_]) < self._zero_thres:
                r += "     -  "
            else:
                r += "{:8.2f}".format(Q_src.at[id_])
            r += "|"
            if dc_bus.loc[dst.at[id_]] and np.abs(Q_dst.at[id_]) < self._zero_thres:
                r += "     -  "
            else:
                r += "{:8.2f}".format(Q_dst.at[id_])
            r += "|"
            r += "{:8.3f}".format(loss_dyn.at[id_])
            r += "|"
            if loss_fix.at[id_] == 0:
                r += "   -   "
            else:
                r += "{:7.3f}".format(loss_fix.at[id_])
            r += "|"
            r += "\n"
            t.append(r)
        t = "".join(t)
        t += "+-------+------+------+----------+----------+--------+--------+--------+-------+\n"
        return t

    def _get_injector_details(self):
        t = ""
        t += "\n+" + "-"*78 + "+\n"
        t += "| INJECTOR RESULT" + " "*62 + "|\n"
        t += "+-------+------+--------------------------+-----------------+------------------+\n"
        t += "|       | Term.|        Injection         |      Cost       |       Type       |\n"
        t += "|       +------+----------+--------+------+--------+--------+                  |\n"
        t += "|       |      |    P     |   Q    |  PF  | for P  | for Q  |                  |\n"
        t += "|  ID   | Bus  |   (MW)   | (Mvar) |      | (k$/h) | (k$/h) |                  |\n"
        t += "+-------+------+----------+--------+------+--------+--------+------------------+\n"
        t = [t]
        scr_injector = self.scenario.injector
        res_injector = self.injector
        terminal = scr_injector['bus']
        dc_bus = self.scenario.bus['type'] == BusType.DC
        S = res_injector['s']
        P = pd.Series(S.real, index=res_injector.index)
        Q = pd.Series(S.imag, index=res_injector.index)
        (P_active, _) = self._check_cap_region_box_active(S, scr_injector['cap'])
        cost_p = res_injector['cost_p']
        cost_q = res_injector['cost_q']
        for id_ in res_injector.index:
            r = "|"
            r += "{:6d}".format(id_)
            r += " |"
            r += "{:5d}".format(terminal.at[id_])
            r += "=" if dc_bus.loc[terminal.at[id_]] else " "
            r += "|"
            r += "*" if P_active.at[id_] else " "
            r += "{:9.2f}".format(P.at[id_])
            r += "|"
            if dc_bus.loc[terminal.at[id_]] and np.abs(Q.at[id_]) < self._zero_thres:
                r += "     -  "
            else:
                r += "{:8.2f}".format(Q.at[id_])
            r += "|"
            if ((dc_bus.loc[terminal.at[id_]] and np.abs(Q.at[id_]) < self._zero_thres)
                    or np.abs(S.at[id_]) < self._zero_thres):
                r += "  -   "
            else:
                r += "{:6.3f}".format(P.at[id_] / np.abs(S.at[id_]))
            r += "|"
            if np.isnan(cost_p.at[id_]):
                r += "    -   "
            else:
                r += "{:8.3f}".format(cost_p.at[id_]/1e3)
            r += "|"
            if np.isnan(cost_q.at[id_]):
                r += "    -   "
            else:
                r += "{:8.3f}".format(cost_q.at[id_]/1e3)
            r += "|"
            r += " {:<16s} ".format(truncate_with_ellipsis(
                                    str(scr_injector.at[id_, 'type']), 16))
            r += "|"
            r += "\n"
            t.append(r)
        t = "".join(t)
        t += "+-------+------+----------+--------+------+--------+--------+------------------+\n"
        return t



#pylint: disable=no-member,logging-format-interpolation
"""
*hynet* optimization client for distributed computation.
"""

import logging
from multiprocessing import Process
from multiprocessing.managers import SyncManager
import queue
import time

import hynet.config as config
from hynet.opf.calc import calc_opf
from hynet.qcqp.problem import QCQP

_log = logging.getLogger(__name__)


def start_optimization_client(server_ip, port=None, authkey=None,
                              num_workers=None, verbose=True):
    """
    Create, connect, and start a *hynet* optimization client.

    Note that this call is blocking until the *hynet* optimization server, to
    which the client is connected, is shut down.

    Parameters
    ----------
    server_ip : str
        IP address of the *hynet* optimization server.
    port : int, optional
        TCP port of the *hynet* optimization server.
    authkey : str, optional
        Authentication key for the *hynet* optimization server.
    num_workers : int, optional
        Number of worker processes that should run in parallel. If more than
        one worker is started, it is recommended to disable the internal
        parallel processing, see ``parallelize`` in ``hynet.config``.
    verbose : bool, optional
        If True (default), some information on activity of the client is
        printed to the standard output.
    """
    if port is None:
        port = config.DISTRIBUTED['default_port']
    if authkey is None:
        authkey = config.DISTRIBUTED['default_authkey']
    if num_workers is None:
        num_workers = config.DISTRIBUTED['default_num_workers']

    manager = _create_client_manager(server_ip, port, authkey.encode('utf-8'))

    workers = []
    for i in range(num_workers):
        workers.append(Process(target=_optimization_worker,
                               args=(i + 1,
                                     manager.get_job_queue(),
                                     manager.get_result_queue(),
                                     verbose)
                               )
                       )
        workers[-1].start()

    if verbose:
        print("Connected client with {0} worker(s) to the hynet server {1}:{2}"
              .format(num_workers, server_ip, port))

    for worker in workers:
        worker.join()


def _create_client_manager(server_ip, port, authkey):
    """
    Return a (connected) manager for a *hynet* optimization client.

    Parameters
    ----------
    server_ip : str
        IP address of the *hynet* optimization server.
    port : int
        TCP port of the *hynet* optimization server.
    authkey : str
        Authentication key for the *hynet* optimization server.

    Returns
    -------
    manager : ClientManager
        Manager object for a *hynet* optimization client.
    """
    class ClientManager(SyncManager):
        """This class manages the synchronization of the queues."""
        pass

    ClientManager.register('get_job_queue')
    ClientManager.register('get_result_queue')

    manager = ClientManager(address=(server_ip, port), authkey=authkey)
    manager.connect()
    _log.info("Connected client to the hynet optimization server {0}:{1}."
              .format(server_ip, port))
    return manager


def _optimization_worker(worker_id, job_queue, result_queue, verbose):
    """
    Run a worker process for *hynet* optimization jobs.

    This worker process repeatedly reads a job from the job queue, performs
    the optimization, and stores the result to the result queue.

    Parameters
    ----------
    worker_id : int
        ID of this worker process.
    job_queue : queue.Queue
        Queue with pending *hynet* optimization jobs.
    result_queue : queue.Queue
        Queue that receives the optimization results.
    verbose : bool, optional
        If True (default), some information on activity of the worker is
        printed to the standard output.
    """
    if verbose:
        def show_info(message):
            """Print a message with the current time stamp."""
            print(time.strftime("%Y-%m-%d %H:%M:%S") + " | " + message)
    else:
        show_info = _log.info

    while True:
        try:
            _log.info("Worker {0}: Waiting for optimization jobs."
                      .format(worker_id))
            job = job_queue.get(block=True, timeout=None)

            show_info("Worker {0}: Starting to process optimization job {1}"
                      .format(worker_id, job.id))
            try:
                if isinstance(job.problem, QCQP):
                    result = job.solver.solve(job.problem)
                else:
                    result = calc_opf(job.problem,
                                      solver=job.solver,
                                      solver_type=job.solver_type)
                out_dict = {job.id: result}
                show_info("Worker {0}: Finished optimization job {1}"
                          .format(worker_id, job.id))
            except Exception as exception:
                out_dict = {job.id: exception}
                show_info("Worker {0}: Job {1} failed: {2}"
                          .format(worker_id, job.id, str(exception)))
            result_queue.put(out_dict)
        except queue.Empty:
            _log.info("Worker {0}: Queue was empty. Resume listening."
                      .format(worker_id))
        except EOFError:
            show_info("Worker {0}: Server was shut down. Exiting..."
                      .format(worker_id))
            return
        except KeyboardInterrupt:
            show_info("Worker {0}: Terminated by the user. Exiting..."
                      .format(worker_id))
            return
        except Exception as exception:
            show_info("Worker {0}: Terminated unexpectedly: {1}"
                      .format(worker_id, str(exception)))
            return



"""
Import of model data into the *hynet* data format.
"""

import logging
import os.path

import numpy as np
import scipy.io
import h5py

from hynet.types_ import (hynet_float_,
                          DBInfoKey,
                          BusType,
                          BranchType,
                          InjectorType,
                          EntityType)
from hynet.data.connection import connect, DBTransaction
from hynet.data.structure import (DBInfo,
                                  DBScenario,
                                  DBBus,
                                  DBBranch,
                                  DBConverter,
                                  DBInjector,
                                  DBCapabilityRegion,
                                  DBSamplePoint,
                                  DBShunt,
                                  DBScenarioLoad,
                                  DBScenarioInjector,
                                  DBScenarioInactivity)

_log = logging.getLogger(__name__)

MAT_FILE_EXTENSION = '.mat'


class BusConstants:
    """Constants for the MATPOWER bus data, see the MATPOWER manual."""

    # Bus matrix column indices
    BUS_I = 0       # Bus number
    BUS_TYPE = 1    # Bus type
    PD = 2          # Real power demand (MW)
    QD = 3          # Reactive power demand (MVAr)
    GS = 4          # Shunt conductance as loss in MW at V = 1.0 p.u.
    BS = 5          # Shunt susceptance as injection in MVAr at V = 1.0 p.u.
    BUS_AREA = 6    # Area number
    VM = 7          # Voltage magnitude (p.u.)
    VA = 8          # Voltage angle (degrees)
    BASE_KV = 9     # Base voltage (kV)
    ZONE = 10       # (Loss) Zone
    VMAX = 11       # Maximum voltage magnitude (p.u.)
    VMIN = 12       # Minimum voltage magnitude (p.u.)

    # Bus types
    PQ = 1          # PQ-bus (fixed variables in power flow calculations)
    PV = 2          # PV-bus (fixed variables in power flow calculations)
    REF = 3         # Reference bus
    NONE = 4        # Isolated/inactive bus

    # OPF result column indices in the bus matrix
    LAM_P = 13      # Lagrange multiplier on real power balance
    LAM_Q = 14      # Lagrange multiplier on reactive power balance
    MU_VMAX = 15    # KKT multiplier on upper voltage limit
    MU_VMIN = 16    # KKT multiplier on lower voltage limit


class BranchConstants:
    """Constants for the MATPOWER branch data, see the MATPOWER manual."""

    # Branch matrix column indices
    F_BUS = 0       # "From" bus number
    T_BUS = 1       # "To" bus number
    BR_R = 2        # Series resistance (p.u.)
    BR_X = 3        # Series reactance (p.u.)
    BR_B = 4        # Total line charging susceptance (p.u.)
    RATE_A = 5      # MVA rating A (long term rating)
    RATE_B = 6      # MVA rating B (short term rating)
    RATE_C = 7      # MVA rating C (emergency rating)
    TAP = 8         # Transformer ratio
    SHIFT = 9       # Transformer phase shift (degrees)
    BR_STATUS = 10  # Status (1 - in service, 0 - out of service)
    ANGMIN = 11     # Minimum angle difference (degrees)
    ANGMAX = 12     # Maximum angle difference (degrees)

    # OPF result column indices in the branch matrix
    PF = 13         # Real power injected at "from" bus end (MW)
    QF = 14         # Reactive power injected at "from" bus end (MVAr)
    PT = 15         # Real power injected at "to" bus end (MW)
    QT = 16         # Reactive power injected at "to" bus end (MVAr)
    MU_SF = 17      # KKT multiplier on MVA limit at "from" bus
    MU_ST = 18      # KKT multiplier on MVA limit at "to" bus
    MU_ANGMIN = 19  # KKT multiplier lower angle difference limit
    MU_ANGMAX = 20  # KKT multiplier upper angle difference limit


class GeneratorConstants:
    """Constants for the MATPOWER generator data, see the MATPOWER manual."""

    # Generator matrix column indices
    GEN_BUS = 0     # Bus number
    PG = 1          # Real power output (MW)
    QG = 2          # Reactive power output (MVAr)
    QMAX = 3        # Maximum reactive power output (MVAr)
    QMIN = 4        # Minimum reactive power output (MVAr)
    VG = 5          # Voltage magnitude set point (p.u.)
    MBASE = 6       # Total MVA base of this machine
    GEN_STATUS = 7  # Status (1 - in service, 0 - out of service)
    PMAX = 8        # Maximum real power output (MW)
    PMIN = 9        # Minimum real power output (MW)
    PC1 = 10        # Lower real power output of PQ capability curve (MW)
    PC2 = 11        # Upper real power output of PQ capability curve (MW)
    QC1MIN = 12     # Minimum reactive power output at PC1 (MVAr)
    QC1MAX = 13     # Maximum reactive power output at PC1 (MVAr)
    QC2MIN = 14     # Minimum reactive power output at PC2 (MVAr)
    QC2MAX = 15     # Maximum reactive power output at PC2 (MVAr)
    RAMP_AGC = 16   # Ramp rate for load following/AGC (MW/min)
    RAMP_10 = 17    # Ramp rate for 10 minute reserves (MW)
    RAMP_30 = 18    # Ramp rate for 30 minute reserves (MW)
    RAMP_Q = 19     # Ramp rate for reactive power (2 sec timescale) (MVAr/min)
    APF = 20        # Area participation factor

    # OPF result column indices in the generator matrix
    MU_PMAX = 21    # KKT multiplier on upper active power upper limit
    MU_PMIN = 22    # KKT multiplier on lower active power upper limit
    MU_QMAX = 23    # KKT multiplier on upper reactive power upper limit
    MU_QMIN = 24    # KKT multiplier on lower reactive power upper limit

    # Generator cost matrix column indices
    MODEL = 0       # Cost model
    STARTUP = 1     # Startup cost in $
    SHUTDOWN = 2    # Shutdown cost in $
    NCOST = 3       # Number of cost coefficients (poly) or sample points (PWL)
    COST = 4        # Start column of coefficient or sample point data

    # Cost model types
    PW_LINEAR = 1   # Piecewise linear cost function
    POLYNOMIAL = 2  # Polynomial cost function


class DCLineConstants:
    """Constants for the MATPOWER DC line data, see the MATPOWER manual."""

    # DC line matrix column indices
    F_BUS = 0       # "From" bus number
    T_BUS = 1       # "To"  bus number
    BR_STATUS = 2   # Status (1 - in service, 0 - out of service)
    PMIN = 9        # Lower limit on MW flow at "from" end
    PMAX = 10       # Upper limit on MW flow at "from" end
    QMINF = 11      # Lower limit on MVAr injection at "from" bus
    QMAXF = 12      # Upper limit on MVAr injection at "from" bus
    QMINT = 13      # Lower limit on MVAr injection at "to" bus
    QMAXT = 14      # Upper limit on MVAr injection at "to" bus
    LOSS0 = 15      # Offset of linear loss function (MW)
    LOSS1 = 16      # Slope of linear loss function

    # OPF result column indices in the DC line matrix
    PF = 3          # Flow at "from" bus in MW
    PT = 4          # Flow at  "to"  bus in MW
    QF = 5          # Injection at "from" bus in MVAr
    QT = 6          # Injection at  "to"  bus in MVAr
    VF = 7          # Voltage set point at "from" bus (p.u.)
    VT = 8          # Voltage set point at "to" bus (p.u.)
    MU_PMIN = 17    # KKT multiplier on lower flow lim at "from" bus
    MU_PMAX = 18    # KKT multiplier on upper flow lim at "from" bus
    MU_QMINF = 19   # KKT multiplier on lower VAr lim at "from" bus
    MU_QMAXF = 20   # KKT multiplier on upper VAr lim at "from" bus
    MU_QMINT = 21   # KKT multiplier on lower VAr lim at "to" bus
    MU_QMAXT = 22   # KKT multiplier on upper VAr lim at "to" bus


def _load_mpc_from_mat_file(input_file):
    """
    Load the MATPOWER case struct from the MAT file.

    SciPy's MAT loading (``scipy.io.loadmat``) only supports the formats
    v4 (Level 1.0), v6, and v7 to 7.2. For later versions, ``h5py`` is
    employed. The data is preprocessed for equal structure among all versions.
    """
    mpc = {}
    try:
        mat = scipy.io.loadmat(input_file)['mpc']
        mpc['baseMVA'] = mat['baseMVA'][0][0][0][0]
        prepare_data = lambda x: x[0][0]
        prepare_dict = prepare_data
        get_keys = lambda x: x.dtype.names
    except NotImplementedError:
        mat = h5py.File(input_file, 'r')['mpc']
        mpc['baseMVA'] = mat['baseMVA'][0][0]

        def fix_empty_matrix(x):
            if x.shape == (2,):
                return np.ndarray(shape=x, dtype=hynet_float_).transpose()
            return x

        prepare_data = lambda x: fix_empty_matrix(x[()]).transpose()
        prepare_dict = lambda x: x
        get_keys = lambda x: x.keys()

    mpc['bus'] = prepare_data(mat['bus'])
    mpc['branch'] = prepare_data(mat['branch'])
    mpc['gen'] = prepare_data(mat['gen'])
    mpc['gencost'] = prepare_data(mat['gencost'])
    try:
        mpc['dcline'] = prepare_data(mat['dcline'])
    except (ValueError, KeyError):
        mpc['dcline'] = np.ndarray((0, DCLineConstants.LOSS1 + 1),
                                   dtype=hynet_float_)

    # --- CUSTOM EXTENSION: Scenario & line length information ---
    try:
        mpc['line_length'] = prepare_data(mat['line_length'])
    except (ValueError, KeyError):
        pass

    if 'line_length' in mpc:
        if mpc['line_length'].shape[0] != mpc['branch'].shape[0]:
            raise ValueError("Line length information is inconsistent.")

    try:
        scenarios = prepare_dict(mat['scenarios'])
    except (ValueError, KeyError):
        pass
    else:
        mpc['scenarios'] = {}
        for scenario in get_keys(scenarios):
            mpc['scenarios'][scenario] = {}
            scenario_data = prepare_dict(scenarios[scenario])
            for field in get_keys(scenario_data):
                mpc['scenarios'][scenario][field] = \
                    prepare_data(scenario_data[field])
    # --- CUSTOM EXTENSION: END ----------------------------------
    return mpc


def _get_gencost_sample_points(cost, p_min, p_max, num_sample_points):
    """Return the PWL function sample points for a MATPOWER ``gencost`` row."""
    cgen = GeneratorConstants
    sample_points = []
    N = int(cost[cgen.NCOST])

    if N < 1:
        return sample_points

    if cost[cgen.MODEL] == cgen.PW_LINEAR:
        for x in range(N):
            sample_points.append((cost[cgen.COST + 2*x],
                                  cost[cgen.COST + 2*x + 1]))
        return sample_points

    if cost[cgen.MODEL] == cgen.POLYNOMIAL:
        coeff = cost[cgen.COST:cgen.COST + N]

        if np.all(coeff == 0):
            return sample_points

        if N == 1:
            # Constant function
            sample_points.append((0, coeff[0]))
            sample_points.append((1, coeff[0]))
            return sample_points

        if np.all(coeff[:-2] == 0):
            # Linear function
            sample_points.append((0, coeff[-1]))
            sample_points.append((1, coeff[-1] + coeff[-2]))
            return sample_points

        # Polynomial of second or higher order
        if p_min > p_max:
            raise ValueError("The test case contains a generator with "
                             "infeasible active power limits: "
                             "PMIN = {:f} > {:f} = PMAX.".format(p_min, p_max))

        if p_min == p_max:
            constant_cost = np.polyval(coeff, p_min)
            sample_points.append((0, constant_cost))
            sample_points.append((1, constant_cost))
            _log.warning("For generators with fixed active power injection, "
                         "the conversion of polynomial cost functions is not "
                         "supported. The cost function is set to the constant "
                         "{:f} that reflects the cost at its operating point."
                         .format(constant_cost))
            return sample_points

        for x in np.linspace(p_min, p_max, num_sample_points):
            sample_points.append((x, np.polyval(coeff, x)))
        return sample_points

    raise ValueError("Invalid cost function type: {:d}".format(cost[cgen.MODEL]))


def import_matpower_test_case(input_file, output_file=None, grid_name=None,
                              description='', num_sample_points=10,
                              res_detection=False):
    """
    Import a MATPOWER test case file into *hynet*'s database format.

    This function imports a MATPOWER test case, which is stored as a MATPOWER
    test case struct ``mpc`` in a MATLAB MAT-file, to a *hynet*'s database. To
    prepare the import, start MATLAB and perform the following two steps:

    1) Load the MATPOWER test case into the variable ``mpc``::

        mpc = loadcase('your_matpower_test_case_file.m');

    2) Save the MATPOWER test case struct ``mpc`` to a MATLAB MAT file::

        save('your_matpower_test_case_file.mat', 'mpc');

    Call this function with the MAT-file as the input.

    Parameters
    ----------
    input_file : str
        MATLAB MAT-file (.mat) with the MATPOWER test case struct ``mpc``.
    output_file : str, optional
        Destination *hynet* grid database file. By default (None), the file
        name is set to the input file name with a ``.db`` extension.
    grid_name : str, optional
        Name of the grid. By default (None), the grid name is set to the input
        file name excluding the extension.
    description : str, optional
        Description of the grid model and, if applicable, copyright and
        licensing information.
    num_sample_points : int, optional
        Number of sample points that shall be used for the conversion of
        polynomial to piecewise linear cost functions (on the interval from
        minimum to maximum output). This setting is a trade-off between an
        accurate representation of the original cost function and the number
        of additional constraints in the OPF problem.
    res_detection : bool, optional
        Detection of the injector type for renewable energy sources (RES),
        which is inactive by default. This scheme is motivated by the German
        grid data, which contains PV- and wind-based injectors with zero
        marginal cost that are arranged in a specific pattern: The injectors
        that connect to the same bus are stored consecutively in the generator
        matrix. If two or more injectors connect to the same bus, then the
        *last* one is wind-based and the *second to the last* is PV-based
        generation if their marginal cost is zero. If this parameter is set to
        True, this RES detection scheme is enabled.

    Returns
    -------
    output_file : str
        Destination *hynet* grid database file name.

    Raises
    ------
    ValueError
    NotImplementedError
    """
    cbus = BusConstants
    cbra = BranchConstants
    cgen = GeneratorConstants
    cdcl = DCLineConstants

    if num_sample_points < 2:
        raise ValueError("Invalid number of sample points (use >= 2).")

    # Check file names
    if not input_file.lower().endswith(MAT_FILE_EXTENSION):
        raise ValueError("The input file is not a MATLAB MAT-file.")

    if output_file is None:
        output_file = "{:}.db".format(input_file[:-len(MAT_FILE_EXTENSION)])

    # Load MATPOWER test case data
    try:
        mpc = _load_mpc_from_mat_file(input_file)
    except Exception as exception:
        raise IOError("Failed to load the MAT file: " + str(exception))

    # Connect to the destination database
    database = connect(output_file)

    if not database.empty:
        raise ValueError("The destination database is not empty.")

    if grid_name is None:
        grid_name = os.path.basename(os.path.splitext(input_file)[0])

    scenario_id = 0             # Scenario ID for the base case
    base_case_inactivity = []   # Track base case inactivity for scenarios
    base_kv = {}                # Track bus base volt. for transformer detection
    with DBTransaction(database) as transaction:
        # General database information
        transaction.add(DBInfo(key=DBInfoKey.GRID_NAME,
                               value=grid_name))
        transaction.add(DBInfo(key=DBInfoKey.BASE_MVA,
                               value=hynet_float_(mpc['baseMVA'])))
        transaction.add(DBInfo(key=DBInfoKey.DESCRIPTION,
                               value=description))

        # Add a default scenario
        transaction.add(DBScenario(id=scenario_id,
                                   time=0.,
                                   name='Base Case',
                                   loss_price=0.,
                                   annotation=''))

        # Import bus data
        if mpc['bus'].shape[0] < 1:
            raise ValueError("The test case does not contain any buses.")

        # Use the area number as zone if there is no zone information available
        ZONE_COLUMN = cbus.ZONE
        if np.all(mpc['bus'][:, cbus.ZONE] == mpc['bus'][0, cbus.ZONE]):
            if np.any(mpc['bus'][:, cbus.BUS_AREA] != mpc['bus'][0, cbus.BUS_AREA]):
                ZONE_COLUMN = cbus.BUS_AREA

        shunt_id = 1
        for bus in mpc['bus']:
            bus_id = int(bus[cbus.BUS_I])
            transaction.add(DBBus(id=bus_id,
                                  type=BusType.AC,
                                  ref=(bus[cbus.BUS_TYPE] == cbus.REF),
                                  base_kv=bus[cbus.BASE_KV],
                                  v_min=bus[cbus.VMIN],
                                  v_max=bus[cbus.VMAX],
                                  zone=int(bus[ZONE_COLUMN]),
                                  annotation=''))
            base_kv[bus_id] = bus[cbus.BASE_KV]

            # Add shunt
            if bus[cbus.GS] != 0 or bus[cbus.BS] != 0:
                transaction.add(DBShunt(id=shunt_id,
                                        bus_id=bus_id,
                                        p=bus[cbus.GS],
                                        q=bus[cbus.BS],
                                        annotation=''))
                shunt_id += 1

            # Add load to the scenario
            if bus[cbus.PD] != 0 or bus[cbus.QD] != 0:
                transaction.add(DBScenarioLoad(scenario_id=scenario_id,
                                               bus_id=bus_id,
                                               p=bus[cbus.PD],
                                               q=bus[cbus.QD]))

            # Is this bus isolated/inactive?
            if bus[cbus.BUS_TYPE] == cbus.NONE:
                inactivity = DBScenarioInactivity(scenario_id=scenario_id,
                                                  entity_type=EntityType.BUS,
                                                  entity_id=bus_id)
                transaction.add(inactivity)
                base_case_inactivity.append(inactivity)

        # Import branch data
        for i, branch in enumerate(mpc['branch']):
            branch_id = i + 1
            src_id = int(branch[cbra.F_BUS])
            dst_id = int(branch[cbra.T_BUS])

            # TRANSFORMER DETECTION: The branch type is set to transformer if
            # the branch connects different voltage levels or if the ratio or
            # phase setting is not neutral.
            #
            # REMARK: MATPOWER includes only a phase shifting transformer at
            # the source bus, which employs a *reciprocal* tap ratio compared
            # to *hynet*.
            ratio_src = 1. if branch[cbra.TAP] == 0 else 1 / branch[cbra.TAP]
            phase_src = -branch[cbra.SHIFT]
            if base_kv[src_id] != base_kv[dst_id] or \
                    ratio_src != 1 or phase_src != 0:
                branch_type = BranchType.TRANSFORMER
            else:
                branch_type = BranchType.LINE

            rating = branch[cbra.RATE_A]
            rating = rating if rating != 0 else None

            angle_min = branch[cbra.ANGMIN]
            angle_max = branch[cbra.ANGMAX]
            if ((angle_min == 0 and angle_max == 0) or
                    (angle_min <= -360 and angle_max >= 360)):
                angle_min = angle_max = None

            # --- CUSTOM EXTENSION: Line length information --------------
            if 'line_length' in mpc:
                line_length = mpc['line_length'][i, 0]
            else:
                line_length = None
            # --- CUSTOM EXTENSION: END ----------------------------------

            transaction.add(DBBranch(id=branch_id,
                                     type=branch_type,
                                     src_id=src_id,
                                     dst_id=dst_id,
                                     r=branch[cbra.BR_R],
                                     x=branch[cbra.BR_X],
                                     b_src=branch[cbra.BR_B] / 2,
                                     b_dst=branch[cbra.BR_B] / 2,
                                     ratio_src=ratio_src,
                                     phase_src=phase_src,
                                     ratio_dst=1.,
                                     phase_dst=0.,
                                     length=line_length,
                                     rating=rating,
                                     angle_min=angle_min,
                                     angle_max=angle_max,
                                     drop_min=None,
                                     drop_max=None,
                                     annotation=''))

            # Is this branch inactive?
            if branch[cbra.BR_STATUS] == 0:
                inactivity = DBScenarioInactivity(scenario_id=scenario_id,
                                                  entity_type=EntityType.BRANCH,
                                                  entity_id=branch_id)
                transaction.add(inactivity)
                base_case_inactivity.append(inactivity)

        # Import DC line data
        # (REMARK: The MATPOWER DC line model is compatible with *hynet*'s
        #  converter model, i.e., DC lines are imported as AC/AC converters.)
        i = 0
        converter_id = 0
        cap_region_id = 1  # Note: This ID counter is also used for inj. cap. reg.
        while i < mpc['dcline'].shape[0]:
            converter_id += 1
            dcline = mpc['dcline'][i]
            p_min = dcline[cdcl.PMIN]
            p_max = dcline[cdcl.PMAX]
            q_min_src = dcline[cdcl.QMINF]
            q_max_src = dcline[cdcl.QMAXF]
            q_min_dst = dcline[cdcl.QMINT]
            q_max_dst = dcline[cdcl.QMAXT]
            loss_fwd = loss_bwd = dcline[cdcl.LOSS1] * 100

            # BIDIRECTIONAL LINES: The MATPOWER data format only supports
            # unidirectional lines and, as a consequence, test cases with
            # HVDC lines typically contain two entries for one line. The
            # following code combines these two entries to a bidirectional
            # HVDC line model.
            if i + 1 < len(mpc['dcline']):
                next_dcline = mpc['dcline'][i + 1]
                if (next_dcline[cdcl.F_BUS] == dcline[cdcl.T_BUS] and
                        next_dcline[cdcl.T_BUS] == dcline[cdcl.F_BUS] and
                        next_dcline[cdcl.BR_STATUS] == dcline[cdcl.BR_STATUS] and
                        next_dcline[cdcl.PMIN] == dcline[cdcl.PMIN] == 0 and
                        next_dcline[cdcl.PMAX] >= 0 and p_max >= 0):
                    p_min = -next_dcline[cdcl.PMAX]
                    q_min_src += next_dcline[cdcl.QMINT]
                    q_max_src += next_dcline[cdcl.QMAXT]
                    q_min_dst += next_dcline[cdcl.QMINF]
                    q_max_dst += next_dcline[cdcl.QMAXF]
                    loss_bwd = next_dcline[cdcl.LOSS1] * 100
                    i += 1  # Skip the second part of this bidirectional line

            # Capability region at the source and destination terminal
            cap_src_id = cap_region_id
            cap_dst_id = cap_region_id + 1
            cap_region_id += 2

            transaction.add(DBCapabilityRegion(id=cap_src_id,
                                               p_min=p_min,
                                               p_max=p_max,
                                               q_min=q_min_src,
                                               q_max=q_max_src,
                                               lt_ofs=None,
                                               lt_slp=None,
                                               rt_ofs=None,
                                               rt_slp=None,
                                               lb_ofs=None,
                                               lb_slp=None,
                                               rb_ofs=None,
                                               rb_slp=None,
                                               annotation=''))

            transaction.add(DBCapabilityRegion(id=cap_dst_id,
                                               p_min=-p_max,  # Flow *into* the
                                               p_max=-p_min,  # converter!
                                               q_min=q_min_dst,
                                               q_max=q_max_dst,
                                               lt_ofs=None,
                                               lt_slp=None,
                                               rt_ofs=None,
                                               rt_slp=None,
                                               lb_ofs=None,
                                               lb_slp=None,
                                               rb_ofs=None,
                                               rb_slp=None,
                                               annotation=''))

            # Add converter
            transaction.add(DBConverter(id=converter_id,
                                        src_id=int(dcline[cdcl.F_BUS]),
                                        dst_id=int(dcline[cdcl.T_BUS]),
                                        cap_src_id=cap_src_id,
                                        cap_dst_id=cap_dst_id,
                                        loss_fwd=loss_fwd,
                                        loss_bwd=loss_bwd,
                                        loss_fix=dcline[cdcl.LOSS0],
                                        annotation=''))

            # Is this DC line inactive?
            if dcline[cdcl.BR_STATUS] == 0:
                inactivity = DBScenarioInactivity(scenario_id=scenario_id,
                                                  entity_type=EntityType.CONVERTER,
                                                  entity_id=converter_id)
                transaction.add(inactivity)
                base_case_inactivity.append(inactivity)
            i += 1

        # Import generator data
        if mpc['gen'].shape[0] != mpc['gencost'].shape[0]:
            raise NotImplementedError("The import of reactive power cost "
                                      "functions is not supported.")

        pwl_function_id = 1
        for i, injector in enumerate(mpc['gen']):
            injector_id = i + 1
            gencost = mpc['gencost'][i]

            # Capability region at the source and destination terminal
            cap_id = cap_region_id
            cap_region_id += 1

            if any(injector[[cgen.PC1, cgen.PC2, cgen.QC1MIN, cgen.QC1MAX,
                             cgen.QC2MIN, cgen.QC2MAX]] != 0):
                _log.warning("The import of the sloped refinements of the "
                             "generator capability are not supported. This "
                             "data is ignored, i.e., only PMIN, PMAX, QMIN, "
                             "QMAX is considered.")

            transaction.add(DBCapabilityRegion(id=cap_id,
                                               p_min=injector[cgen.PMIN],
                                               p_max=injector[cgen.PMAX],
                                               q_min=injector[cgen.QMIN],
                                               q_max=injector[cgen.QMAX],
                                               lt_ofs=None,
                                               lt_slp=None,
                                               rt_ofs=None,
                                               rt_slp=None,
                                               lb_ofs=None,
                                               lb_slp=None,
                                               rb_ofs=None,
                                               rb_slp=None,
                                               annotation=''))

            # PWL cost function for active power
            sample_points = _get_gencost_sample_points(gencost,
                                                       injector[cgen.PMIN],
                                                       injector[cgen.PMAX],
                                                       num_sample_points)
            if sample_points:
                cost_id = pwl_function_id
                pwl_function_id += 1
                for x, y in sample_points:
                    transaction.add(DBSamplePoint(id=cost_id, x=x, y=y))
            else:
                cost_id = None

            # INJECTOR TYPE: By default, it is set to conventional generation.
            # If and only if the generator offers exclusively reactive power at
            # zero cost, then it is considered as shunt compensation.
            terminal = int(injector[cgen.GEN_BUS])
            injector_type = InjectorType.CONVENTIONAL
            if cost_id is None and \
                    all(injector[[cgen.PMIN, cgen.PMAX]] == 0) and \
                    any(injector[[cgen.QMIN, cgen.QMAX]] != 0):
                injector_type = InjectorType.COMPENSATION
            # --- CUSTOM EXTENSION: RES injector type detection ----------
            if res_detection and cost_id is None:
                num_gen = mpc['gen'].shape[0]
                if i + 1 < num_gen:
                    next_terminal = int(mpc['gen'][i + 1, cgen.GEN_BUS])
                    if next_terminal == terminal:
                        if (i + 2 == num_gen or
                                int(mpc['gen'][i + 2, cgen.GEN_BUS] != terminal)):
                            next_cost = _get_gencost_sample_points(
                                mpc['gencost'][i + 1],
                                mpc['gen'][i + 1, cgen.PMIN],
                                mpc['gen'][i + 1, cgen.PMAX],
                                num_sample_points)
                            if not next_cost:
                                injector_type = InjectorType.PV
                if i >= 1:
                    prev_terminal = int(mpc['gen'][i - 1][cgen.GEN_BUS])
                    if prev_terminal == terminal:
                        if (i + 1 == num_gen or
                                int(mpc['gen'][i + 1][cgen.GEN_BUS] != terminal)):
                            prev_cost = _get_gencost_sample_points(
                                mpc['gencost'][i - 1],
                                mpc['gen'][i - 1, cgen.PMIN],
                                mpc['gen'][i - 1, cgen.PMAX],
                                num_sample_points)
                            if not prev_cost:
                                injector_type = InjectorType.WIND
            # --- CUSTOM EXTENSION: END ----------------------------------

            # Add converter
            if injector[cgen.MBASE] not in [0., mpc['baseMVA']]:
                # REMARK: The MVA base of the machine does not seem to be
                # utilized at any point in the MATPOWER code and, in the vast
                # majority of test cases, it is set to the test case MVA base.
                _log.debug("The generator MVA base does not match the MVA "
                           "base of the test case.")

            ramp = injector[cgen.RAMP_AGC] * 60  # Convert from MW/min to MW/h
            transaction.add(DBInjector(id=injector_id,
                                       type=injector_type,
                                       bus_id=terminal,
                                       cap_id=cap_id,
                                       cost_p_id=cost_id,
                                       cost_q_id=None,
                                       cost_start=gencost[cgen.STARTUP],
                                       cost_stop=gencost[cgen.SHUTDOWN],
                                       ramp_up=ramp,
                                       ramp_down=ramp,
                                       min_up=None,
                                       min_down=None,
                                       energy_min=None,
                                       energy_max=None,
                                       annotation=''))

            # Is this generator inactive?
            if injector[cgen.GEN_STATUS] <= 0:
                inactivity = DBScenarioInactivity(scenario_id=scenario_id,
                                                  entity_type=EntityType.INJECTOR,
                                                  entity_id=injector_id)
                transaction.add(inactivity)
                base_case_inactivity.append(inactivity)

        # If there is no scenario data, we're done...
        if 'scenarios' not in mpc:
            return output_file

        # --- CUSTOM EXTENSION: Scenario information -----------------
        for scenario_name in mpc['scenarios']:
            scenario_data = mpc['scenarios'][scenario_name]
            scenario_title = scenario_name.replace('_', ' ').title()

            load_p = scenario_data['load_p']
            load_q = scenario_data['load_q']
            gen_pmin = scenario_data['gen_pmin']
            gen_pmax = scenario_data['gen_pmax']
            gen_qmin = scenario_data['gen_qmin']
            gen_qmax = scenario_data['gen_qmax']

            hours = load_p.shape[1]
            if not all([x.shape[1] == hours for x in [load_p, load_q,
                                                      gen_pmin, gen_pmax,
                                                      gen_qmin, gen_qmax]]):
                raise ValueError("The scenario data is inconsistent.")

            for hour in range(hours):
                scenario_id += 1
                with DBTransaction(database) as transaction:
                    transaction.add(DBScenario(id=scenario_id,
                                               time=hour,
                                               name=scenario_title,
                                               loss_price=0.,
                                               annotation=''))

                    for i, bus_id in enumerate(mpc['bus'][:, cbus.BUS_I]):
                        if load_p[i, hour] == 0 and load_q[i, hour] == 0:
                            continue
                        transaction.add(DBScenarioLoad(scenario_id=scenario_id,
                                                       bus_id=int(bus_id),
                                                       p=load_p[i, hour],
                                                       q=load_q[i, hour]))

                    for i, injector in enumerate(mpc['gen']):
                        injector_id = i + 1
                        if (gen_pmin[i, hour] == injector[cgen.PMIN] and
                                gen_pmax[i, hour] == injector[cgen.PMAX] and
                                gen_qmin[i, hour] == injector[cgen.QMIN] and
                                gen_qmax[i, hour] == injector[cgen.QMAX]):
                            continue
                        transaction.add(DBScenarioInjector(scenario_id=scenario_id,
                                                           injector_id=injector_id,
                                                           p_min=gen_pmin[i, hour],
                                                           p_max=gen_pmax[i, hour],
                                                           q_min=gen_qmin[i, hour],
                                                           q_max=gen_qmax[i, hour],
                                                           cost_scaling=1.))

                    for inactivity in base_case_inactivity:
                        transaction.add(
                            DBScenarioInactivity(scenario_id=scenario_id,
                                                 entity_type=inactivity.entity_type,
                                                 entity_id=inactivity.entity_id))
        # --- CUSTOM EXTENSION: END ----------------------------------
    return output_file



"""
PICOS-based solver for the *hynet*-specific QCQP problem (SD and SOC relaxation).
"""

import logging
import abc

import picos as pic
import cvxopt as cvx
import numpy as np

from hynet.types_ import hynet_float_, SolverType, SolverStatus
from hynet.utilities.base import Timer
from hynet.qcqp.solver import SolverInterface
from hynet.qcqp.problem import QCQPPoint
from hynet.qcqp.result import QCQPResult

_log = logging.getLogger(__name__)


def parse_status(status):
    """Parse the PICOS status into a ``SolverStatus``."""
    if status.upper() == 'OPTIMAL':
        return SolverStatus.SOLVED
    if status.upper() in ['NEAR OPTIMAL', 'INACCURATE']:
        return SolverStatus.INACCURATE
    if status.upper() in ['INFEASIBLE', 'PRIMAL INFEASIBLE', 'DUAL INFEASIBLE']:
        return SolverStatus.INFEASIBLE
    return SolverStatus.FAILED


def scipy2cvxopt(matrix):
    """
    Return the SciPy sparse matrix as a CVXOPT sparse matrix.

    PICOS only accepts matrices in CVXOPT's format. This function offers the
    corresponding type conversion.

    Parameters
    ----------
    matrix : .hynet_sparse_
        Matrix that shall be converted.
    Returns
    -------
    cvx.spmatrix
        Input matrix in CVXOPT's sparse format.
    """
    A = matrix.tocoo()
    return cvx.spmatrix(A.data.tolist(), A.row.tolist(), A.col.tolist(), A.shape)


class SolverBase(SolverInterface):
    """Base class for PICOS-based solvers for a relaxation of the QCQP."""

    def __init__(self, solver_name='cvxopt', **kwds):
        super().__init__(**kwds)
        self.solver_name = solver_name

        if self.solver_name.upper() == 'CVXOPT':
            if 'tol' not in self.param:
                self.param['tol'] = 1e-7
            # REMARK: By default, the parameters are set as reltol = 10 * tol,
            # abstol = tol, and feastol = tol.
            if 'reltol' not in self.param:
                self.param['reltol'] = 10 * self.param['tol']
            if 'abstol' not in self.param:
                self.param['abstol'] = 10 * self.param['tol']
            if 'feastol' not in self.param:
                self.param['feastol'] = 100 * self.param['tol']
            if 'maxit' not in self.param:
                self.param['maxit'] = 500

    @staticmethod
    @abc.abstractmethod
    def _add_voltage_matrix_constraints(prob, qcqp, x):
        """Add the SOC or PSD constraint(s) on the ``v_tld``-part of ``x``."""

    def solve(self, qcqp):
        """
        Solve the given QCQP via a relaxation using PICOS.

        The relaxation-specific constraints on the bus voltage matrix are
        added via the (abstract) method ``_add_voltage_matrix_constraints``,
        which must be implemented in a derived class.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the relaxed problem associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                ...and SOC or PSD constraint(s) on v_tld-part of x
        #
        # For the following code, I used the PICOS documentation provided at
        # http://picos.zib.de/api.html

        timer = Timer()
        # ------------------ generate data --------------------------
        (c, A, b, B, d, x_lb, x_ub) = qcqp.get_vectorization()

        # -------------------- size of the data ---------------------
        dim_x = c.shape[0]

        #  ----------- convert to CVXOPT sparse matrices ------------
        # The PICOS interface uses CVXOPT to deal with matrices
        c = scipy2cvxopt(c)
        A = scipy2cvxopt(A)
        B = scipy2cvxopt(B)

        # ---- convert the constants into params of the problem -----
        c = pic.new_param('c', c)
        A = pic.new_param('A', A)
        b = pic.new_param('b', b)
        B = pic.new_param('B', B)
        d = pic.new_param('d', d)

        # ----- create a problem and the optimization variable ------
        prob = pic.Problem()
        x = prob.add_variable('x', dim_x)

        # ------------------- add the constraints -------------------
        # REMARK: In PICOS, the operators < and > implement <= and >=, resp.
        crt_Ax_b = prob.add_constraint(A * x == b, ret=True)
        crt_Bx_d = prob.add_constraint(B * x < d, ret=True)

        # Add the specified lower and upper bounds
        #
        # REMARK: The bound constraints are not implemented with ``lower`` and
        # ``upper`` of ``add_variable`` as they (apparently) do not provide
        # the corresponding dual variables. Furthermore, as PICOS (apparently)
        # does not support variable indexing with index arrays, the cumbersome
        # approach below is pursued.
        def create_selection_matrix(idx):
            return cvx.spmatrix(np.ones(len(idx)),    # Coefficients
                                np.arange(len(idx)),  # Row indices
                                idx,                  # Column indices
                                (len(idx), dim_x))    # Size

        idx_lb = np.where(~np.isnan(x_lb))[0]
        idx_ub = np.where(~np.isnan(x_ub))[0]

        S_lb = pic.new_param('S_lb', create_selection_matrix(idx_lb))
        S_ub = pic.new_param('S_ub', create_selection_matrix(idx_ub))

        x_lb = pic.new_param('x_lb', x_lb[idx_lb])
        x_ub = pic.new_param('x_ub', x_ub[idx_ub])

        crt_lb = prob.add_constraint(S_lb * x > x_lb, ret=True)
        crt_ub = prob.add_constraint(S_ub * x < x_ub, ret=True)

        # Add relaxation-specific constraints on the voltage matrix
        # (This method is overridden appropriately in derived classes)
        self._add_voltage_matrix_constraints(prob, qcqp, x)

        # -------------------- set the objective --------------------
        prob.set_objective('min', c|x)

        _log.debug(self.type.name + " ~ Modeling ({:.3f} sec.)"
                   .format(timer.interval()))

        # ----------- optimization and solution recovery ------------
        prob.solve(solver=self.solver_name,
                   verbose=1 if self.verbose else 0,
                   **self.param)
        solver_time = timer.interval()
        _log.debug(self.type.name + " ~ Solver ({:.3f} sec.)"
                   .format(solver_time))

        status = parse_status(prob.status)

        if status in [SolverStatus.SOLVED, SolverStatus.INACCURATE]:
            # Convert the PICOS variable object into a NumPy array
            x = np.asarray(x.value).reshape(dim_x)

            (V, f, s, z) = qcqp.split_vectorization_optimizer(x)
            v = self.rank1approx(V, qcqp.edges, qcqp.roots)
            optimizer = QCQPPoint(v, f, s, z)

            dv_eq = np.asarray(crt_Ax_b.dual).reshape(-1)
            dv_ineq = np.asarray(crt_Bx_d.dual).reshape(-1)

            dv_lb = np.nan * np.ones(dim_x, dtype=hynet_float_)
            dv_lb[idx_lb] = np.asarray(crt_lb.dual).reshape(-1)
            dv_lb = qcqp.split_vectorization_bound_dual(dv_lb)

            dv_ub = np.nan * np.ones(dim_x, dtype=hynet_float_)
            dv_ub[idx_ub] = np.asarray(crt_ub.dual).reshape(-1)
            dv_ub = qcqp.split_vectorization_bound_dual(dv_ub)

            optimal_value = prob.obj_value()

            result = QCQPResult(qcqp, self, status, solver_time,
                                optimizer=optimizer,
                                V=V,
                                optimal_value=optimal_value,
                                dv_lb=dv_lb,
                                dv_ub=dv_ub,
                                dv_eq=dv_eq,
                                dv_ineq=dv_ineq)
        else:
            result = QCQPResult(qcqp, self, status, solver_time)

        _log.debug(self.type.name + " ~ Result creation ({:.3f} sec.)"
                   .format(timer.interval()))
        _log.debug(self.type.name + " ~ Total time for solver ({:.3f} sec.)"
                   .format(timer.total()))
        return result


class SOCRSolver(SolverBase):
    """
    PICOS-based solver for a second-order cone relaxation of the QCQP.

    For information on solver-specific parameters, see e.g. [1]_. To retrieve a
    list of available solvers, use

    >>> import picos
    >>> picos.tools.available_solvers()

    By default, CVXOPT is selected. For appropriate performance in OPF
    calculations, its default parameters are adapted. Access the ``param``
    attribute of an object for their inspection.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.

    References
    ----------
    .. [1] http://picos.zib.de/problem.html
    """

    @property
    def name(self):
        return "PICOS + " + self.solver_name.upper()

    @property
    def type(self):
        return SolverType.SOCR

    def solve(self, qcqp):  # pylint: disable=useless-super-delegation
        """
        Solve the given QCQP via its second-order cone relaxation using PICOS.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the SOCR associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                   ...and SOC constraints on v_tld-part of x
        return super().solve(qcqp)

    @staticmethod
    def _add_voltage_matrix_constraints(prob, qcqp, x):
        """Add the SOC constraints on the ``v_tld``-part of ``x``."""
        # Second-order cone constraints on V (cf. QCQP.get_vectorization)
        #
        # PSD constraint on 2x2 principal submatrix =>
        #  (a) diagonal is nonnegative -> ensured by LB on voltage mag.
        #  (b) determinant is nonnegative -> rotated SOC:
        #
        #               V_ij * conj(V_ij) <= V_ii * V_jj
        #                             <=>
        #          real(V_ij)^2 + imag(V_ij)^2 <= V_ii * V_jj
        #
        #      where i is the source bus and j is the destination bus
        #      of the considered edge. The latter is implemented below.
        #
        # Note that PICOS supports two types of SOC constraints:
        #  (1) Lorentz cone: abs(x) <= t;
        #  (2) Rotated SOC: abs(x)**2 <= t*u; t >=0
        dim_v = qcqp.dim_v
        K = qcqp.num_edges
        (e_src, e_dst) = qcqp.edges

        prob.add_list_of_constraints(
            [abs(x[(dim_v + k):(dim_v + k + K + 1):K])**2
             < x[int(e_src[k])] * x[int(e_dst[k])]
             for k in range(K)], 'k', '[K]')


class SDRSolver(SolverBase):
    """
    PICOS-based solver for a semidefinite relaxation of the QCQP.

    For information on solver-specific parameters, see e.g. [1]_. To retrieve a
    list of available solvers, use

    >>> import picos
    >>> picos.tools.available_solvers()

    By default, CVXOPT is selected. For appropriate performance in OPF
    calculations, its default parameters are adapted. Access the ``param``
    attribute of an object for their inspection.

    See Also
    --------
    hynet.qcqp.solver.SolverInterface : Abstract base class for QCQP solvers.

    References
    ----------
    .. [1] http://picos.zib.de/problem.html
    """

    @property
    def name(self):
        return "PICOS + " + self.solver_name.upper()

    @property
    def type(self):
        return SolverType.SDR

    def solve(self, qcqp):  # pylint: disable=useless-super-delegation
        """
        Solve the given QCQP via its semidefinite relaxation using PICOS.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the SDR associated with the QCQP.
        """
        # We solve the following problem:
        #
        #  min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
        #   x                   ...and a PSD constraint on the v_tld-part of x
        return super().solve(qcqp)

    @staticmethod
    def _add_voltage_matrix_constraints(prob, qcqp, x):
        """Add the PSD constraint on the ``v_tld``-part of ``x``."""
        # PSD constraint on V
        #
        # Consider the splitting of the Hermitian matrix V into its
        # real and imaginary part, i.e., V = V_r + j*V_i. Then,
        #
        #       V is PSD    <=>    V_cr = [ V_r  -V_i ] is PSD
        #                                 [ V_i   V_r ]
        #
        # This code implements the real-valued equivalent of V >= 0,
        # i.e., V_cr >= 0. To this end, a semidefinite matrix variable
        # V_cr is added and all relevant elements are equated to v_tld.
        dim_v = qcqp.dim_v
        K = qcqp.num_edges
        (e_src, e_dst) = qcqp.edges

        if np.any(e_src >= e_dst):
            raise RuntimeError("For simplicity, this solver assumes "
                               "that the source node index is smaller "
                               "than the destination node index in the "
                               "sparsity specification. At the time of "
                               "implementation, this was the default.")

        V_tld = prob.add_variable('V_tld', (2*dim_v, 2*dim_v), 'symmetric')

        prob.add_constraint(V_tld >> 0)  # PSD

        for k in range(K):
            i = int(e_src[k])
            j = int(e_dst[k])
            prob.add_constraint(V_tld[i, j] == x[dim_v + k])
            prob.add_constraint(V_tld[i + dim_v, j + dim_v] == x[dim_v + k])
            prob.add_constraint(V_tld[i + dim_v, j] == x[dim_v + k + K])
            prob.add_constraint(V_tld[i, j + dim_v] == -x[dim_v + k + K])
        for i in range(dim_v):
            prob.add_constraint(V_tld[i, i] == x[i])
            prob.add_constraint(V_tld[i + dim_v, i + dim_v] == x[i])
            prob.add_constraint(V_tld[i + dim_v, i] == 0)



import os
from setuptools import setup, find_packages


def read(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name),
              encoding='utf-8') as file:
        return file.read()


setup(
    name='hynet',
    version='1.0.3',  # Update the version in __init__.py and conf.py as well
    author='Matthias Hotz',
    author_email='matthias.hotz@tum.de',
    description='An optimal power flow framework for hybrid AC/DC power systems.',
    long_description=read('README.md'),
    license='BSD 3-clause license',
    keywords='hybrid AC/DC power systems optimal power flow convex relaxation',
    url='http://www.msv.ei.tum.de/',
    packages=find_packages(exclude=['tests.*', 'tests', 'expansion']),
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: BSD License',
    ],
    python_requires='>=3.5',
    install_requires=[
        'sqlalchemy>=1.2.3',
        'numpy>=1.14.0',
        'scipy>=1.0.0',
        'pandas>=0.23.0',
        'matplotlib>=2.1.1',
        'h5py>=2.8.0'
    ],
    extras_require={
        'graph': ['networkx>=2.1'],
        'test': ['pytest>=3.5.0', 'pytest-cov', 'pylint>=1.8.4', 'pylint-exit',
                 'PICOS>=1.1.2', 'cvxopt>=1.2.1'],
        'doc': ['Sphinx>=1.7.8', 'sphinxcontrib-napoleon>=0.6.1',
                'numpydoc>=0.8.0', 'm2r>=0.1.14']
    }
)



"""
Support for OPF regression tests.
"""

import logging
from collections import namedtuple

import numpy as np

_log = logging.getLogger(__name__)


class OPFVerificationData(namedtuple('OPFVerificationData',
                                     ['OPTIMAL_VALUE',
                                      'VOLTAGES',
                                      'BRANCH_FLOW_SRC',
                                      'CONVERTER_FLOW_SRC',
                                      'DISPATCH',
                                      'DV_BALANCE_P',
                                      'DV_CAP_P_MAX',
                                      'DV_V_MAX'])):
    """
    OPF solution verification data for regression tests.

    This class contains the minimum amount of reference data to properly verify
    the outcome of an OPF calculation. It was introduced to compactly embed the
    reference solution for installation testing in *hynet*'s code - it's not
    too pleasing, but simple and robust ;)
    """
    @staticmethod
    def create(result):
        """
        Return a OPF verification data object for the given OPF result.

        Parameters
        ----------
        result : hynet.opf.result.OPFResult
            OPF result for which the OPF verification data shall be created.

        Returns
        -------
        reference : OPFVerificationData
            OPF verification data.
        """
        return OPFVerificationData(
            OPTIMAL_VALUE=result.optimal_value,
            VOLTAGES=result.bus['v'].values,
            BRANCH_FLOW_SRC=result.branch['s_src'].values,
            CONVERTER_FLOW_SRC=(result.converter['p_src']
                                - 1j * result.converter['q_src']).values,
            DISPATCH=result.injector['s'].values,
            DV_BALANCE_P=result.bus['dv_bal_p'].values,
            DV_CAP_P_MAX=result.injector['dv_cap_p_max'].values,
            DV_V_MAX=result.bus['dv_v_max'].values
        )


def _rel_mae(value, reference):
    """Return the relative mean absolute error."""
    return np.sum(np.abs(value - reference)) / np.sum(np.abs(reference))


def verify_opf_result(result, reference, tolerance):
    """
    Verify an OPF result versus a reference solution.

    The OPF result is verified by comparing the OPF verification data to a
    reference solution by calculating the respective relative mean absolute
    errors and checking them against a tolerance threshold. As the bus voltage
    magnitudes are less rigid (in particular w.r.t. relaxations), the error
    threshold is relaxed for those comparisons.

    Parameters
    ----------
    result : OPFVerificationData
        OPF solution that shall be verified.
    reference : OPFVerificationData
        Reference OPF solution.
    tolerance : .hynet_float_
        Tolerance on the relative mean absolute error of the individual result
        vectors.

    Raises
    ------
    ValueError
        In case any mismatch beyond the tolerance is detected.
    """
    MESSAGE_END = " not match the expected result."

    # REMARK: The checks are formulated in a negated form to detect NaNs as well.
    _log.debug("Relative mean absolute errors to reference OPF solution:")

    rel_err = _rel_mae(result.OPTIMAL_VALUE, reference.OPTIMAL_VALUE)
    _log.debug(" - Optimal objective value: {:.6e}".format(rel_err))
    if not rel_err <= tolerance:
        raise ValueError("The optimal objective does" + MESSAGE_END)

    rel_err = _rel_mae(np.abs(result.VOLTAGES), np.abs(reference.VOLTAGES))
    _log.debug(" - Bus voltage magnitudes: {:.6e}".format(rel_err))
    if not rel_err <= 10 * tolerance:  # Be tolerant for the relaxations... :)
        raise ValueError("The optimal bus voltage magnitudes do" + MESSAGE_END)

    rel_err = _rel_mae(np.angle(result.VOLTAGES), np.angle(reference.VOLTAGES))
    _log.debug(" - Bus voltage angles: {:.6e}".format(rel_err))
    if not rel_err <= tolerance:
        raise ValueError("The optimal bus voltage angles do" + MESSAGE_END)

    rel_err = _rel_mae(result.BRANCH_FLOW_SRC, reference.BRANCH_FLOW_SRC)
    _log.debug(" - Branch flows (at source bus): {:.6e}".format(rel_err))
    if not rel_err <= tolerance:
        raise ValueError("The branch flows do" + MESSAGE_END)

    rel_err = _rel_mae(result.CONVERTER_FLOW_SRC, reference.CONVERTER_FLOW_SRC)
    _log.debug(" - Converter flows (at source bus): {:.6e}".format(rel_err))
    if not rel_err <= tolerance:
        raise ValueError("The converter flows do" + MESSAGE_END)

    rel_err = _rel_mae(result.DISPATCH, reference.DISPATCH)
    _log.debug(" - Injector dispatch: {:.6e}".format(rel_err))
    if not rel_err <= tolerance:
        raise ValueError("The injector dispatch does" + MESSAGE_END)

    rel_err = _rel_mae(result.DV_BALANCE_P, reference.DV_BALANCE_P)
    _log.debug(" - Active power balance dual variables: {:.6e}".format(rel_err))
    if not rel_err <= tolerance:
        raise ValueError("The active power balance dual variables do" +
                         MESSAGE_END)

    rel_err = _rel_mae(result.DV_CAP_P_MAX, reference.DV_CAP_P_MAX)
    _log.debug(" - Injector active power UB dual variables: {:.6e}"
               .format(rel_err))
    if not rel_err <= tolerance:
        raise ValueError("The injector active power UB dual variables do" +
                         MESSAGE_END)

    rel_err = _rel_mae(result.DV_V_MAX, reference.DV_V_MAX)
    _log.debug(" - Voltage UB dual variables: {:.6e}".format(rel_err))
    if not rel_err <= 10 * tolerance:
        raise ValueError("The voltage UB dual variables do" + MESSAGE_END)



#pylint: disable=C0330,bad-whitespace
"""
Artificial system for OPF regression tests.

This testing data is embedded as code to obtain simple and robust access to it.
"""

import numpy as np

from hynet.types_ import (hynet_id_,
                          hynet_float_,
                          hynet_complex_,
                          BusType,
                          BranchType,
                          InjectorType)
from hynet.scenario.representation import Scenario
from hynet.scenario.capability import HalfSpace, CapRegion, ConverterCapRegion
from hynet.scenario.cost import PWLFunction
from hynet.test.regression import OPFVerificationData


# =============================================================================
# ============================= TEST SYSTEM SETUP =============================
# =============================================================================

TEST_SYSTEM = Scenario()
TEST_SYSTEM.id = 0
TEST_SYSTEM.name = 'default'
TEST_SYSTEM.time = 25.26  # 1d 01:15:36
TEST_SYSTEM.database_uri = 'test_system.db'
TEST_SYSTEM.grid_name = 'Artificial System for Testing Purposes'
TEST_SYSTEM.base_mva = 100
TEST_SYSTEM.loss_price = 0.25
TEST_SYSTEM.description = ("Manually synthesized system with several special "
                           "cases for testing purposes.")
TEST_SYSTEM.annotation = ''


# === BUS DATA ===

TEST_SYSTEM.bus['id'] = np.arange(1, 12, dtype=hynet_id_)
TEST_SYSTEM.bus['type'] = np.array([BusType.AC] * 5 +
                                   [BusType.DC] * 3 +
                                   [BusType.AC] * 2 +
                                   [BusType.DC], dtype=BusType)
TEST_SYSTEM.bus['ref'] = np.array([False] * 2 +
                                  [True] +
                                  [False] * 2 +
                                  [True] +
                                  [False] * 2 +
                                  [True] +
                                  [False] * 2, dtype=bool)
TEST_SYSTEM.bus['base_kv'] = np.array([380] * 5 +
                                      [585] * 3 +
                                      [220] * 2 +
                                      [10], dtype=hynet_float_)
TEST_SYSTEM.bus['y_tld'] = np.array([0.05j] +
                                    [0.03+0.05j] +
                                    [0j] * 9, dtype=hynet_complex_)
TEST_SYSTEM.bus['load'] = np.array([0j,
                                    300 + 100j,
                                    300 + 100j,
                                    400 + 129j,
                                    0j,
                                    0j,
                                    0j,
                                    0j,
                                    38 + 13j,
                                    0j,
                                    3.3], dtype=hynet_complex_)
TEST_SYSTEM.bus['v_min'] = np.array([0.9] * 9 +
                                    [1.025] +
                                    [0.9], dtype=hynet_float_)
TEST_SYSTEM.bus['v_max'] = np.array([1.1] * 10 +
                                    [1.2], dtype=hynet_float_)
TEST_SYSTEM.bus['zone'] = np.array([None] * 11, dtype=object)
TEST_SYSTEM.bus['annotation'] = np.array([''] * 11, dtype=str)
TEST_SYSTEM.bus.set_index('id', inplace=True)


# === BRANCH DATA ===

TEST_SYSTEM.branch['id'] = np.array([1, 2, 3, 4, 7, 6, 5, -1, -2, -3],
                                    dtype=hynet_id_)
TEST_SYSTEM.branch['type'] = np.array([BranchType.LINE] +
                                      [BranchType.TRANSFORMER] * 2 +
                                      [BranchType.LINE] * 7, dtype=BranchType)
TEST_SYSTEM.branch['src'] = np.array([1, 1, 5, 2, 6, 8, 10, 9, 1, 6],
                                     dtype=hynet_id_)
TEST_SYSTEM.branch['dst'] = np.array([2, 4, 1, 3, 7, 7, 9, 10, 4, 7],
                                     dtype=hynet_id_)
TEST_SYSTEM.branch['z_bar'] = np.array([0.00281 + 0.02810j,
                                        0.00304 + 0.03040j,
                                        0.00064 + 0.00640j,
                                        0.00108 + 0.01080j,
                                        0.00297 + 0.00000j,
                                        0.00297 + 0.00000j,
                                        0.00281 + 0.02810j,
                                        0.00281 + 0.02810j,
                                        0.00304 + 0.03040j,
                                        0.00297 + 0.00000j],
                                       dtype=hynet_complex_)
TEST_SYSTEM.branch['y_src'] = np.array([0.00356j,
                                        0.00329j,
                                        0.01563j,
                                        0.00926j,
                                        0.00000j,
                                        0.00000j,
                                        0.00000j,
                                        0.00000j,
                                        0.00329j,
                                        0.00000j], dtype=hynet_complex_)
TEST_SYSTEM.branch['y_dst'] = TEST_SYSTEM.branch['y_src'].copy()
TEST_SYSTEM.branch['rho_src'] = np.array([1 + 0j] +
                                         [0.95 + 0.01j] +
                                         [1 + 0j] * 4 +
                                         [1.05 + 0.02j] +
                                         [1 + 0j] +
                                         [0.95 + 0.01j] +
                                         [1 + 0j], dtype=hynet_complex_)
TEST_SYSTEM.branch['rho_dst'] = np.array([1 + 0j] * 7 +
                                         [1.05+0.02j] +
                                         [1 + 0j] * 2, dtype=hynet_complex_)
TEST_SYSTEM.branch['length'] = np.array([np.nan] * 10, dtype=hynet_float_)
TEST_SYSTEM.branch['rating'] = np.array([400,
                                         250,
                                         500,
                                         None,
                                         250,
                                         125,
                                         50,
                                         50,
                                         250,
                                         250], dtype=hynet_float_)
TEST_SYSTEM.branch['angle_min'] = np.array([-4,
                                            -50,
                                            None,
                                            -50] +
                                           [np.nan] * 6, dtype=hynet_float_)
TEST_SYSTEM.branch['angle_max'] = np.array([None,
                                            -0.82,
                                            5,
                                            50,
                                            None,
                                            None,
                                            1] +
                                           [np.nan] * 3, dtype=hynet_float_)
TEST_SYSTEM.branch['drop_min'] = np.array([-10,
                                           -7,
                                           None,
                                           -10,
                                           None,
                                           -0.4] +
                                          [np.nan] * 4, dtype=hynet_float_)
TEST_SYSTEM.branch['drop_max'] = np.array([10,
                                           None,
                                           10,
                                           10,
                                           0,
                                           None,
                                           4.815] +
                                          [np.nan] * 3, dtype=hynet_float_)
TEST_SYSTEM.branch['annotation'] = np.array([''] * 10, dtype=str)
TEST_SYSTEM.branch.set_index('id', inplace=True)


# === CONVERTER DATA ===

TEST_SYSTEM.converter['id'] = np.arange(1, 6, dtype=hynet_id_)
TEST_SYSTEM.converter['src'] = np.array([3, 7, 5, 1, 8], dtype=hynet_id_)
TEST_SYSTEM.converter['dst'] = np.array([6, 4, 8, 2, 7], dtype=hynet_id_)
TEST_SYSTEM.converter['cap_src'] = \
    np.array([ConverterCapRegion([-50, 50], [-30, 30],
                                 rt=HalfSpace(0.1, -0.05)),
              ConverterCapRegion([-500, 500]),
              ConverterCapRegion([-240, 240], [-3, 1]),
              ConverterCapRegion([-10, 10]),
              ConverterCapRegion([-5, 5])], dtype=ConverterCapRegion)
TEST_SYSTEM.converter['cap_dst'] = \
    np.array([ConverterCapRegion([-500, 500]),
              ConverterCapRegion([-500, 500], [-2, 2]),
              ConverterCapRegion([-139, 240]),
              ConverterCapRegion([-10, 10]),
              ConverterCapRegion([-5, 5])], dtype=ConverterCapRegion)
TEST_SYSTEM.converter['loss_fwd'] = np.array([1.0] * 5, dtype=hynet_float_)
TEST_SYSTEM.converter['loss_bwd'] = np.array([2.0] * 5, dtype=hynet_float_)
TEST_SYSTEM.converter['loss_fix'] = np.array([0.75] +
                                             [0.0] * 4, dtype=hynet_float_)
TEST_SYSTEM.converter['annotation'] = np.array([''] * 5, dtype=str)
TEST_SYSTEM.converter.set_index('id', inplace=True)


# === INJECTOR DATA ===

TEST_SYSTEM.injector['id'] = np.arange(1, 7, dtype=hynet_id_)
TEST_SYSTEM.injector['bus'] = np.array([1, 3, 5, 7, 10, 11], dtype=hynet_id_)
TEST_SYSTEM.injector['type'] = np.array([InjectorType.CONVENTIONAL,
                                         InjectorType.RENEWABLE,
                                         InjectorType.PROSUMER,
                                         InjectorType.LOAD,
                                         InjectorType.WIND,
                                         InjectorType.PV], dtype=InjectorType)
TEST_SYSTEM.injector['cap'] = \
    np.array([CapRegion([0, 210], [-155, 155],
                        lt=HalfSpace(0.5, 0.3)),
              CapRegion([0, 520], [-390, 390]),
              CapRegion([0, 600], [-450, 450],
                        rt=HalfSpace(0.5, -1),
                        rb=HalfSpace(0.5, 1)),
              CapRegion([0, 50]),
              CapRegion([0, 100], [-25, 25]),
              CapRegion([0, 10])], dtype=CapRegion)
TEST_SYSTEM.injector['cost_p'] = \
    np.array([PWLFunction(((0, 1), (0, 14))),
              PWLFunction(((0, 1), (0, 30))),
              PWLFunction(((0, 1), (0, 10))),
              PWLFunction(((0, 1), (0, 15))),
              PWLFunction(((0, 1), (0, 58))),
              None], dtype=PWLFunction)
TEST_SYSTEM.injector['cost_q'] = \
    np.array([PWLFunction(((-1, 0, 1), (1, 0, 1))),
              PWLFunction(((-1, 0, 1), (1, 0, 1))),
              PWLFunction(((-1, 0, 1), (1, 0, 1))),
              None,
              PWLFunction(((-1, 0, 1), (1, 0, 1))),
              None], dtype=PWLFunction)
TEST_SYSTEM.injector['cost_start'] = np.array([0.0] * 6, dtype=hynet_float_)
TEST_SYSTEM.injector['cost_stop'] = np.array([0.0] * 6, dtype=hynet_float_)
TEST_SYSTEM.injector['annotation'] = np.array([''] * 6, dtype=str)
TEST_SYSTEM.injector.set_index('id', inplace=True)


# =============================================================================
# ============================ REFERENCE SOLUTION =============================
# =============================================================================

# The reference solution below was created using the IPOPT solver with
# tol=1e-7 using the following procedure:
#
# >>> import hynet as ht
# >>> from hynet.test.system import TEST_SYSTEM
# >>> from hynet.test.regression import OPFVerificationData
# >>> result = ht.calc_opf(TEST_SYSTEM, solver=ht.solver.ipopt.QCQPSolver())
# >>> print(OPFVerificationData.create(result))

REFERENCE_SOLUTION = OPFVerificationData(
    OPTIMAL_VALUE=19655.22273549465,
    VOLTAGES=\
        np.array([1.0919586 +0.07750003j,
                  1.08905516+0.00113417j,
                  1.1       +0.j        ,
                  1.01644898+0.0575343j ,
                  1.09504455+0.10429503j,
                  1.09666635+0.j        ,
                  1.09628691+0.j        ,
                  1.0999994 +0.j        ,
                  1.09998981+0.j        ,
                  1.04935731-0.01552334j,
                  1.02769582+0.j        ]),
    BRANCH_FLOW_SRC=\
        np.array([296.22710879  +2.29512619j,
                   93.18572865 +66.49986007j,
                  458.49715249 +48.72515685j,
                    0.51120455-111.50356669j,
                   14.01099107  +0.j        ,
                  137.4995977   +0.j        ,
                   19.0093649   +6.59364905j,
                  -19.          -6.5j       ,
                   93.18572865 +66.49986007j,
                   14.01099107  +0.j        ]),
    CONVERTER_FLOW_SRC=\
        np.array([ 28.30503321-4.0847257j ,
                  216.53322113+0.j        ,
                  140.40404289-0.99997736j,
                    9.99999974+0.j        ,
                    1.50040251+0.j        ]),
    DISPATCH=\
        np.array([ 35.22687002 +88.06805944j,
                  328.65482601+206.3100792j ,
                  598.90119538 +47.72517948j,
                   49.99999946  +0.j        ,
                   38.01872981 +13.1872981j ,
                    3.3         +0.j        ]),
    DV_BALANCE_P=\
        np.array([ 1.12417151e+01,
                   3.02360457e+01,
                   3.00000001e+01,
                   3.06866128e+01,
                   1.00000217e+01,
                   3.03560606e+01,
                   3.03772467e+01,
                   3.00709779e+01,
                   5.80602080e+01,
                   5.79999998e+01,
                  -3.63471732e-06]),
    DV_CAP_P_MAX=\
        np.array([1.35240259e-07,
                  1.23527354e-07,
                  2.15108598e-05,
                  1.53772471e+01,
                  3.81346868e-07,
                  3.52781494e-06]),
    DV_V_MAX=\
        np.array([2.03408114e-03,
                  9.86592092e-04,
                  1.09514608e+04,
                  1.36215692e-04,
                  7.16133310e+04,
                  3.22771939e-03,
                  2.89837525e-03,
                  1.76416102e+01,
                  1.05373417e+00,
                  2.17661720e-04,
                  6.15778506e-05])
)



"""
Solver interface for a *hynet*-specific QCQP.
"""

import logging
import abc

import numpy as np

from hynet.utilities.graph import get_graph_components
from hynet.utilities.rank1approx import rank1approx_via_least_squares

_log = logging.getLogger(__name__)


class SolverInterface(abc.ABC):
    """
    Abstract base class for QCQP solvers.

    Derived classes implement a solver for the *hynet*-specific quadratically
    constrained quadratic program specified by an object of the class ``QCQP``.
    A solver may solve the nonconvex QCQP, its semidefinite relaxation (SDR),
    or its second-order cone relaxation (SOCR).

    Parameters
    ----------
    verbose : bool, optional
        If ``True``, the logging information of the solver is printed to the
        standard output.
    param : dict[str, object], optional
        Dictionary of parameters (``{'parameter_name': value}``) to modify the
        solver's default settings.
    rank1approx : function, optional
        Function for the rank-1 approximation of a partial Hermitian matrix.
        This method is used in relaxation-based solvers.

    See Also
    --------
    hynet.utilities.rank1approx : Rank-1 approximation functions.
    """

    def __init__(self, verbose=False, param=None,
                 rank1approx=rank1approx_via_least_squares):
        self.verbose = verbose
        # REMARK: It looks like everyone, at some point, has to stumble across
        # the highly unintuitive behavior of mutable defaults...for me it was
        # when the parameters of some solver popped up in another one. So be
        # aware and check out http://effbot.org/zone/default-values.htm ;)
        self.param = {} if param is None else param
        self.rank1approx = rank1approx

    def __str__(self):
        return self.name + ' (' + self.type.value + ')'

    @property
    @abc.abstractmethod
    def name(self):
        """Return the name of the solver."""

    @property
    @abc.abstractmethod
    def type(self):
        """Return the type of the solver as a SolverType."""

    @abc.abstractmethod
    def solve(self, qcqp):
        """
        Solve the given QCQP and return a ``QCQPResult`` object.

        Parameters
        ----------
        qcqp : QCQP
            Specification of the quadratically-constrained quadratic problem.

        Returns
        -------
        result : QCQPResult
            Solution of the QCQP.
        """

    @staticmethod
    def adjust_absolute_phase(qcqp, v):
        """
        Return ``v`` with absolute phase adjustment according to the root nodes.

        Due to the quadratic form in ``v``, the solution of the QCQP is
        ambiguous w.r.t. a common phase shift of the elements in ``v``
        associated with a connected component in the sparsity pattern. This
        method adjusts the absolute phase such that it is zero at the root
        node of the respective components.

        *Remark:* This method should be used in *QCQP solvers*. In SDR and SOCR
        solvers, the rank-1 approximation takes care of the absolute phase
        adjustment.

        Parameters
        ----------
        v : numpy.ndarray[.hynet_complex_]
            Optimizer ``v`` for the solved QCQP.

        Returns
        -------
        v : numpy.ndarray[.hynet_complex_]
            Optimizer ``v`` with absolute phase adjustment.
        """
        v_ = np.array(v)
        components = get_graph_components(np.arange(qcqp.dim_v), qcqp.edges,
                                          roots=qcqp.roots)
        for component in components:
            root = component[np.isin(component, qcqp.roots)]
            if len(root) == 0:  # pylint: disable=len-as-condition
                root = component[0]
            else:
                if len(root) > 1:
                    _log.warning("Sparsity pattern has ambiguous roots.")
                root = root[0]
            v_[component] *= np.exp(-1j * np.angle(v_[root]))
        return v_



"""
Manage *hynet*'s database connections.
"""

import logging
import textwrap

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from hynet.data.structure import (SCHEMA_VERSION,
                                  Base,
                                  DBInfo,
                                  DBBus,
                                  DBBranch,
                                  DBConverter,
                                  DBInjector,
                                  DBShunt,
                                  DBScenario)
from hynet.types_ import DBInfoKey

_log = logging.getLogger(__name__)

DIALECT_PREFIX_SEPARATOR = ':///'
DESCRIPTION_LINE_LENGTH = 75


def connect(database_uri):
    """
    Return a connection to the specified *hynet* grid database.

    Parameters
    ----------
    database_uri : str
        URI or file name of the *hynet* grid database.

    Returns
    -------
    database : DBConnection
        Connection to the *hynet* grid database.
    """
    if DIALECT_PREFIX_SEPARATOR not in database_uri:
        database_uri = 'sqlite' + DIALECT_PREFIX_SEPARATOR + database_uri
    return DBConnection(database_uri)


class DBConnection:
    """
    Manager for a *hynet* grid database connection.

    See Also
    --------
    DBTransaction
    """

    def __init__(self, database_uri):
        """
        Establish a connection to the specified *hynet* grid database.

        Parameters
        ----------
        database_uri: str
            URI of the *hynet* grid database.
        """
        # Create an engine and bind it to the metadata and session maker
        self.database_uri = database_uri
        self.engine = create_engine(database_uri)

        Base.metadata.bind = self.engine
        Base.metadata.create_all(self.engine)  # Create tables if they're absent

        self._session_maker = sessionmaker(bind=self.engine)

        # Check database version or, if empty, initialize it
        with DBTransaction(self) as transaction:
            try:
                version = transaction.query(DBInfo).filter(
                    DBInfo.key == DBInfoKey.VERSION).one().value
                if float(version) < float(SCHEMA_VERSION):
                    raise IOError("Deprecated database version '"
                                  + version.value + "' (expected '"
                                  + SCHEMA_VERSION + "' or higher)")
            except NoResultFound:
                transaction.add(DBInfo(key=DBInfoKey.VERSION,
                                       value=SCHEMA_VERSION))
        _log.debug("Connected to database '{:s}'".format(self.database_uri))

    @property
    def empty(self):
        """Return True if the database does not contain grid information."""
        is_empty = True
        with DBTransaction(self) as transaction:
            for table in [DBBus, DBBranch, DBConverter,
                          DBInjector, DBShunt, DBScenario]:
                if transaction.query(table).first() is not None:
                    is_empty = False
                    break
        return is_empty

    @property
    def version(self):
        """Return the *hynet* grid database format version of this database."""
        return self.get_setting(DBInfoKey.VERSION)

    @property
    def grid_name(self):
        """Return the name of the grid in this database."""
        return '' if self.empty else self.get_setting(DBInfoKey.GRID_NAME)

    @property
    def description(self):
        """Return the description of this database."""
        if self.empty:
            return ''
        return textwrap.fill(self.get_setting(DBInfoKey.DESCRIPTION),
                             width=DESCRIPTION_LINE_LENGTH)

    def get_setting(self, key):
        """
        Return the database setting for the specified key.

        Parameters
        ----------
        key: DBInfoKey
            The key for which the value shall be retrieved.

        Returns
        -------
        value : str
            The value associated with the provided key.

        Raises
        ------
        ValueError
            If the setting was not found.
        """
        with DBTransaction(self) as transaction:
            try:
                setting = transaction.query(DBInfo)\
                            .filter(DBInfo.key == key).one().value
            except NoResultFound:
                raise ValueError("Setting '{:s}' was not found.".format(key))
        return setting

    def start_session(self):
        """
        Return a new database session as an SQLAlchemy Session object.

        **Remark:** This function is for internal use.
        """
        return self._session_maker()


class DBTransaction:
    """
    Database transaction that is automatically committed at the exit block.
    """

    def __init__(self, database):
        """
        Create a database transaction.

        Parameters
        ----------
        database: DBConnection
            Connection to the *hynet* grid database.
        """
        self._session = database.start_session()

    def __enter__(self):
        """
        Initiate the transaction.

        Returns
        -------
        DBTransaction
            ``self``.
        """
        return self

    def __exit__(self, type_, value, traceback):
        """
        Commit the transaction.

        Attempts to commit the current transaction. In case that the commit
        fails, a roll back of the transaction is attempted.

        Parameters
        ----------
        type_ : Exception class or None
            In case of an exception, class of the exception. None otherwise.
        value : Exception instance or None
            In case of an exception, the exception object. None otherwise.
        traceback : object or None
            In case of an exception, object with traceback info. None otherwise.
        """
        if type_ is None:
            self._session.commit()
        else:
            self._session.rollback()
        self._session.close()
        self._session = None
        return False  # In case of an exception, reraise it

    def add(self, object_):
        """Add an object to the session."""
        self._session.add(object_)

    def add_all(self, collection):
        """Add a collection of objects to the session."""
        self._session.add_all(collection)

    def update(self, object_):
        """Update the state of the object in the session."""
        self._session.merge(object_)

    def delete(self, object_):
        """Mark the object as deleted in the session."""
        self._session.delete(object_)

    def delete_all(self, collection):
        """Mark a collection of objects as deleted"""
        for object_ in collection:
            self.delete(object_)

    def query(self, object_type):
        """Return a new SQLAlchemy Query object for this session."""
        return self._session.query(object_type)

    def execute(self, query):
        """Execute the SQL expression construct or string statement."""
        return self._session.execute(query)
